from typing import Union

import brevitas.nn as bnn

from elasticai.creator.layers import QConv1d, QConv2d
from elasticai.creator.brevitas.translation_functions.translation_function_tools import (
    set_quantizers,
)


def translate_conv(layer: Union[QConv1d, QConv2d]) -> dict:
    """
    translates the arguments of a qtorch convolutional layer into the arguments of a brevitas convolutional layer
    the padding mode of the qtorch layer is translated to the padding type of the brevitas layer, brevitas does not support all padding modes
    the arguments conversion of the quantizers is done in an extra file because it is also used by the linear layer
    Args:
        layer (Conv): current convolutional layer
    Returns:
        a dictionary with all arguments for the brevitas convolutional layer
    """
    args = {
        "in_channels": layer.in_channels,
        "out_channels": layer.out_channels,
        "kernel_size": layer.kernel_size,
        "stride": layer.stride,
        "padding": layer.padding,
        "padding_type": "standard",
        "dilation": layer.dilation,
        "groups": layer.groups,
        "bias": layer.bias is not None,
    }

    if layer.padding_mode == "zeros":
        if isinstance(layer.padding, str):
            if layer.padding == "same":
                args["padding_type"] = "same"
                args["padding"] = 0
            else:
                raise NotImplementedError(f"Padding {layer.padding} not implemented.")
    else:
        raise NotImplementedError(f"Padding mode {layer.padding} not implemented.")

    set_quantizers(layer, args)

    return args


def translate_conv1d(layer: QConv1d) -> bnn.QuantConv1d:
    """
    translates a 1d qtorch convolutional layer into a 1d brevitas convolutional layer
    Args:
        layer (Conv): qtorch layer
    Returns:
        brevitas layer
    """
    args = translate_conv(layer=layer)
    return bnn.QuantConv1d(**args)


def translate_conv2d(layer: QConv2d) -> bnn.QuantConv2d:
    """
    translates a 2d qtorch convolutional layer into a 2d brevitas convolutional layer
    Args:
        layer (Conv): qtorch layer
    Returns:
        brevitas layer
    """
    args = translate_conv(layer=layer)
    return bnn.QuantConv2d(**args)



import unittest

from elasticai.creator.brevitas.brevitas_model_comparison import (
    BrevitasModelMatcher,
)


def dummy_model():
    yield from []


class DummyLayer:
    def __init__(self, *parameters):
        self._parameters = [DummyParameter(x) for x in parameters]

    def parameters(self):
        yield from self._parameters


class DummyParameter:
    def __init__(self, value):
        self.value = value

    def equal(self, other):
        if other is None:
            return False
        return self.value == other.value


class BrevitasModelComparisonTest(unittest.TestCase):
    def test_one_model_and_None_raise_Typeerror(self):
        self.assertRaises(
            TypeError, BrevitasModelMatcher.check_equality, None, dummy_model()
        )
        self.assertRaises(
            TypeError, BrevitasModelMatcher.check_equality, dummy_model(), None
        )

    def test_empty_models_are_equal(self):
        fake_model_1 = dummy_model()
        fake_model_2 = dummy_model()
        equality = BrevitasModelMatcher.check_equality(fake_model_1, fake_model_2)
        self.assertTrue(equality)

    def test_models_with_one_layer_are_equal(self):
        fake_model_1 = [DummyLayer()]
        fake_model_2 = [DummyLayer()]
        equality = BrevitasModelMatcher.check_equality(fake_model_1, fake_model_2)
        self.assertTrue(equality)

    def test_models_with_one_layer_are_unequal(self):
        fake_model_1 = [1]
        fake_model_2 = ["bla"]
        equality = BrevitasModelMatcher.check_equality(fake_model_1, fake_model_2)
        self.assertFalse(equality)

    def test_models_with_different_number_of_layers(self):
        dummy_model_1 = [DummyLayer(), "bla"]
        dummy_model_2 = [DummyLayer()]
        equality = BrevitasModelMatcher.check_equality(dummy_model_1, dummy_model_2)
        self.assertFalse(equality)

    def test_non_iterable_models_throw_exception(self):
        dummy_model_1 = 1
        dummy_model_2 = 2
        self.assertRaises(
            TypeError, BrevitasModelMatcher.check_equality, dummy_model_1, dummy_model_2
        )

    def test_models_with_same_layer_types_but_different_parameters_are_unequal(self):
        dummy_model_1 = [DummyLayer(1)]
        dummy_model_2 = [DummyLayer()]
        equality = BrevitasModelMatcher.check_equality(dummy_model_1, dummy_model_2)
        self.assertFalse(equality)

    def test_models_with_same_layer_types_but_different_parameters_are_unequal_exchanged(
        self,
    ):
        dummy_model_1 = [DummyLayer()]
        dummy_model_2 = [DummyLayer(1)]
        equality = BrevitasModelMatcher.check_equality(dummy_model_1, dummy_model_2)
        self.assertFalse(equality)

    def test_same_models_with_completely_same_layers_are_equal(self):
        dummy_model_1 = [DummyLayer(1)]
        dummy_model_2 = [DummyLayer(1)]
        equality = BrevitasModelMatcher.check_equality(dummy_model_1, dummy_model_2)
        self.assertTrue(equality)


if __name__ == "__main__":
    unittest.main()



from torch.nn import Module


def translate_layer(layer: Module) -> Module:
    """
    translation of a layer which stays the same, the same layer is just returned
    Args:
        layer (Layer): PyTorch layer
    Returns:
        input layer
    """
    return layer



from brevitas.quant.base import (
    SignedBinaryClampedConst,
    NarrowIntQuant,
    PerTensorConstScaling2bit,
)
from brevitas.core.function_wrapper import InplaceTensorClampSte, TensorClamp
from brevitas.quant.solver import WeightQuantSolver, ActQuantSolver, BiasQuantSolver


class BinaryWeights(SignedBinaryClampedConst, WeightQuantSolver):
    """
    Quantize weights the same way as QTorch Binarize quantizer
    """

    tensor_clamp_impl = InplaceTensorClampSte
    scaling_const = 1


class BinaryBias(SignedBinaryClampedConst, BiasQuantSolver):
    """
    Quantize bias the same way as QTorch Binarize quantizer
    """

    tensor_clamp_impl = InplaceTensorClampSte
    scaling_const = 1
    requires_input_scale = False
    requires_input_bit_width = False


class BinaryActivation(SignedBinaryClampedConst, ActQuantSolver):
    """
    Quantize values the same way as QTorch Binarize quantizer
    """

    tensor_clamp_impl = TensorClamp
    min_val = -1.0
    max_val = 1.0


class TernaryWeights(NarrowIntQuant, PerTensorConstScaling2bit, WeightQuantSolver):
    """
    Quantize weights the same way as QTorch Ternarize quantizer
    with a zero_window_width of 0.5 and without widening_factor
    """

    tensor_clamp_impl = InplaceTensorClampSte
    scaling_const = 1


class TernaryBias(NarrowIntQuant, PerTensorConstScaling2bit, BiasQuantSolver):
    """
    Quantize bias the same way as QTorch Ternarize quantizer
    with a zero_window_width of 0.5 and without widening_factor
    """

    tensor_clamp_impl = InplaceTensorClampSte
    scaling_const = 1
    requires_input_scale = False
    requires_input_bit_width = False


class TernaryActivation(NarrowIntQuant, PerTensorConstScaling2bit, ActQuantSolver):
    """
    Quantize values the same way as QTorch Ternarize quantizer
    with a zero_window_width of 0.5 and without widening_factor
    """

    tensor_clamp_impl = TensorClamp
    min_val = -1.0
    max_val = 1.0



from unittest import TestCase
from typing import Callable, Iterable, Any
from itertools import zip_longest
from torch.nn import Module
from torch import Tensor


class BrevitasModelMatcher:
    """
    self implemented Test case for comparing two brevitas models
    """

    @classmethod
    def check_equality(cls, first_model: Module, second_model: Module) -> bool:
        """
        returns if two brevitas model are equal
        Args:
            first_model (Model): brevitas model
            second_model (Model): brevitas model
        Returns:
            equality boolean
        """
        if first_model is None or second_model is None:
            raise TypeError
        if first_model == second_model:
            return True
        return cls._condition_function_applies_to_all_elements_in_iterables(
            cls._layers_are_equal, first_model, second_model
        )

    @classmethod
    def _condition_function_applies_to_all_elements_in_iterables(
        cls,
        condition_function: Callable[[Any, Any], bool],
        first_iterable: Iterable[Any],
        second_iterable: Iterable[Any],
    ) -> bool:
        """
        applies a condition function to two iterables, e.g. to the layers or to the parameters
        Args:
            condition_function (Callable[[Any, Any], bool]): function
            first_iterable (Iterable[Any]): iterable of first model, e.g. the layers
            second_iterable (Iterable[Any]): iterable of second model, e.g. the layers
        Returns:
            bool, if condition function is met
        """
        for first_item, second_item in zip_longest(first_iterable, second_iterable):
            if not condition_function(first_item, second_item):
                return False
        return True

    @classmethod
    def _layers_are_equal(
        cls, first_model_layer: Module, second_model_layer: Module
    ) -> bool:
        """
        checks if two layers of two models are equal
        Args:
            first_model_layer (Layer): layer of the first model
            second_model_layer (Layer): layer of the second model
        Returns:
            equality boolean of the layers
        """
        if type(first_model_layer) == type(second_model_layer):
            first_model_parameters = cls._get_parameters_or_empty_list(
                first_model_layer
            )
            second_model_parameters = cls._get_parameters_or_empty_list(
                second_model_layer
            )
            return cls._condition_function_applies_to_all_elements_in_iterables(
                cls._parameters_are_equal,
                first_model_parameters,
                second_model_parameters,
            )
        return False

    @classmethod
    def _parameters_are_equal(
        cls, first_model_parameter: Any, second_model_parameter: Any
    ) -> bool:
        """
        compares a parameter of two equal layers in the two models
        Args:
            first_model_parameter (Tensor): parameter of layer of first model
            second_model_parameter (Tensor): parameter of layer of first model
        Returns:
            equality boolean of the parameters
        """
        return first_model_parameter is not None and first_model_parameter.equal(
            second_model_parameter
        )

    @staticmethod
    def _has_callable_parameters(layer: Module) -> bool:
        """
        checks if the current layer has a callable parameters attribute
        Args:
            layer (Layer): brevitas layer
        Returns:
            if layer has a callable parameters attribute
        """
        return hasattr(layer, "parameters") and callable(layer.parameters)

    @classmethod
    def _get_parameters_or_empty_list(cls, layer: Module) -> Iterable:
        """
        returns parameters of an layer if it has parameters
        Args:
            layer (Layer): brevitas layer
        Returns:
            list of parameters or empty list
        """
        if cls._has_callable_parameters(layer):
            return layer.parameters()
        return []


class BrevitasModelComparisonTestCase(TestCase):
    def assertModelEqual(self, expected: Module, other: Module) -> None:
        """
        compares two brevitas model for equality
        Args:
            expected (Model): brevitas model
            other (Model): brevitas model
        """
        self.assertTrue(BrevitasModelMatcher.check_equality(expected, other))



import unittest
from elasticai.creator.vhdl.generator.specific_testprocess_strings import (
    get_test_process_for_one_input_results_in_one_output_string,
    get_test_process_for_multiple_input_results_in_one_output_string,
)


class SpecificTestProcessStringsTest(unittest.TestCase):
    def test_generate_test_process_for_one_input(self) -> None:
        expected_test_process_lines = [
            "        test_input <=  to_signed(-1281,16);",
            "        wait for 1*clk_period;",
            "        report \"The value of 'test_output' is \" & integer'image(to_integer(unsigned(test_output)));",
            '        assert test_output=0 report "The test case -1281 fail" severity failure;',
            "",
            "        test_input <=  to_signed(-1000,16);",
            "        wait for 1*clk_period;",
            "        report \"The value of 'test_output' is \" & integer'image(to_integer(unsigned(test_output)));",
            '        assert test_output=4 report "The test case -1000 fail" severity failure;',
            "",
            "        test_input <=  to_signed(-500,16);",
            "        wait for 1*clk_period;",
            "        report \"The value of 'test_output' is \" & integer'image(to_integer(unsigned(test_output)));",
            '        assert test_output=28 report "The test case -500 fail" severity failure;',
        ]
        test_process_string = (
            get_test_process_for_one_input_results_in_one_output_string(
                inputs=[-1281, -1000, -500],
                outputs=[0, 4, 28],
                input_name="test_input",
                output_name="test_output",
            )
        )
        for i in range(len(expected_test_process_lines)):
            self.assertEqual(
                expected_test_process_lines[i], test_process_string.splitlines()[i]
            )

    def test_generate_test_process_raises_error_when_called_with_different_inputs_and_outputs_lenghts(
        self,
    ) -> None:
        self.assertRaises(
            TypeError,
            get_test_process_for_one_input_results_in_one_output_string,
            [1],
            [],
        )

    def test_generate_test_process_for_multiple_inputs(self) -> None:
        expected_test_process_lines = [
            '        X_MEM <= (x"0013",x"0000",x"0010",x"0013",x"000c",x"0005",x"0005",x"0013",x"0004",x"0002");',
            '        W_MEM <= (x"0011",x"0018",x"0000",x"000d",x"0014",x"000f",x"0012",x"0007",x"0017",x"0012");',
            '        b <= x"008a";',
            "",
            "        reset <= '1';",
            "        wait for 2*clk_period;",
            "        wait until clock = '0';",
            "        reset <= '0';",
            "        wait until ready = '1';",
            "",
            "        report \"expected output is 142, value of 'y' is \" & integer'image(to_integer(signed(y)));",
            '        assert y = 142 report "The 0. test case fail" severity error;',
            "        reset <= '1';",
            "        wait for 1*clk_period;",
            "",
            '        X_MEM <= (x"0014",x"000d",x"0017",x"0008",x"0002",x"0007",x"0002",x"0015",x"0001",x"0010");',
            '        W_MEM <= (x"000e",x"0014",x"0005",x"0015",x"0009",x"0013",x"0007",x"0016",x"0008",x"0004");',
            '        b <= x"0064";',
            "",
            "        reset <= '1';",
            "        wait for 2*clk_period;",
            "        wait until clock = '0';",
            "        reset <= '0';",
            "        wait until ready = '1';",
            "",
            "        report \"expected output is 105, value of 'y' is \" & integer'image(to_integer(signed(y)));",
            '        assert y = 105 report "The 1. test case fail" severity error;',
            "        reset <= '1';",
            "        wait for 1*clk_period;",
            "",
            '        X_MEM <= (x"000f",x"0017",x"000d",x"000f",x"0001",x"0009",x"0002",x"0007",x"0008",x"0013");',
            '        W_MEM <= (x"0001",x"000a",x"0008",x"0010",x"0008",x"0001",x"0016",x"0013",x"0016",x"000a");',
            '        b <= x"009b";',
            "",
            "        reset <= '1';",
            "        wait for 2*clk_period;",
            "        wait until clock = '0';",
            "        reset <= '0';",
            "        wait until ready = '1';",
            "",
            "        report \"expected output is 159, value of 'y' is \" & integer'image(to_integer(signed(y)));",
            '        assert y = 159 report "The 2. test case fail" severity error;',
            "        reset <= '1';",
            "        wait for 1*clk_period;",
            "",
            '        X_MEM <= (x"000c",x"0007",x"0001",x"0019",x"0008",x"000c",x"0019",x"000b",x"0008",x"000d");',
            '        W_MEM <= (x"000e",x"0015",x"0001",x"000b",x"0014",x"0012",x"000f",x"0000",x"0008",x"000e");',
            '        b <= x"004c";',
            "",
            "        reset <= '1';",
            "        wait for 2*clk_period;",
            "        wait until clock = '0';",
            "        reset <= '0';",
            "        wait until ready = '1';",
            "",
            "        report \"expected output is 82, value of 'y' is \" & integer'image(to_integer(signed(y)));",
            '        assert y = 82 report "The 3. test case fail" severity error;',
            "        reset <= '1';",
            "        wait for 1*clk_period;",
            "",
            '        X_MEM <= (x"0005",x"0013",x"0002",x"0013",x"000c",x"000f",x"0003",x"0004",x"0010",x"0001");',
            '        W_MEM <= (x"0006",x"000d",x"0005",x"0009",x"0017",x"0017",x"000e",x"000d",x"0000",x"0019");',
            '        b <= x"0092";',
            "",
            "        reset <= '1';",
            "        wait for 2*clk_period;",
            "        wait until clock = '0';",
            "        reset <= '0';",
            "        wait until ready = '1';",
            "",
            "        report \"expected output is 150, value of 'y' is \" & integer'image(to_integer(signed(y)));",
            '        assert y = 150 report "The 4. test case fail" severity error;',
            "        reset <= '1';",
            "        wait for 1*clk_period;",
        ]
        inputs = [
            {
                "X_MEM": '(x"0013",x"0000",x"0010",x"0013",x"000c",x"0005",x"0005",x"0013",x"0004",x"0002")',
                "W_MEM": '(x"0011",x"0018",x"0000",x"000d",x"0014",x"000f",x"0012",x"0007",x"0017",x"0012")',
                "b": 'x"008a"',
            },
            {
                "X_MEM": '(x"0014",x"000d",x"0017",x"0008",x"0002",x"0007",x"0002",x"0015",x"0001",x"0010")',
                "W_MEM": '(x"000e",x"0014",x"0005",x"0015",x"0009",x"0013",x"0007",x"0016",x"0008",x"0004")',
                "b": 'x"0064"',
            },
            {
                "X_MEM": '(x"000f",x"0017",x"000d",x"000f",x"0001",x"0009",x"0002",x"0007",x"0008",x"0013")',
                "W_MEM": '(x"0001",x"000a",x"0008",x"0010",x"0008",x"0001",x"0016",x"0013",x"0016",x"000a")',
                "b": 'x"009b"',
            },
            {
                "X_MEM": '(x"000c",x"0007",x"0001",x"0019",x"0008",x"000c",x"0019",x"000b",x"0008",x"000d")',
                "W_MEM": '(x"000e",x"0015",x"0001",x"000b",x"0014",x"0012",x"000f",x"0000",x"0008",x"000e")',
                "b": 'x"004c"',
            },
            {
                "X_MEM": '(x"0005",x"0013",x"0002",x"0013",x"000c",x"000f",x"0003",x"0004",x"0010",x"0001")',
                "W_MEM": '(x"0006",x"000d",x"0005",x"0009",x"0017",x"0017",x"000e",x"000d",x"0000",x"0019")',
                "b": 'x"0092"',
            },
        ]
        # expected signal, as test reference output signal
        outputs = [142, 105, 159, 82, 150]
        test_process_string = (
            get_test_process_for_multiple_input_results_in_one_output_string(
                inputs=inputs, outputs=outputs, output_name="y"
            )
        )
        for i in range(len(expected_test_process_lines)):
            self.assertEqual(
                expected_test_process_lines[i], test_process_string.splitlines()[i]
            )

    def test_generate_test_process_multiple_inputs_raises_error_when_called_with_different_inputs_and_outputs_lenghts(
        self,
    ) -> None:
        self.assertRaises(
            TypeError,
            get_test_process_for_multiple_input_results_in_one_output_string,
            [1],
            [],
            "y",
        )



import unittest
import torch
from torch import nn

import brevitas.nn as bnn

from elasticai.creator.brevitas.translation_functions.conv import (
    translate_conv1d,
)
from elasticai.creator.integrationTests.conv_params_comparison import (
    ConvTest,
)
from elasticai.creator.layers import QConv1d, Binarize, Ternarize
import elasticai.creator.brevitas.brevitas_quantizers as bquant


# When groups == in_channels and out_channels == K * in_channels, where K is a positive integer, this operation is also known as a “depthwise convolution”.
class DepthwiseConv1dTest(ConvTest):
    @staticmethod
    def create_qtorch_conv_layers(
        in_channel, quantizer, padding=0, padding_mode="zeros"
    ):
        return QConv1d(
            in_channels=in_channel,
            out_channels=10,
            kernel_size=10,
            stride=1,
            padding=padding,
            dilation=1,
            groups=1,
            bias=True,
            padding_mode=padding_mode,
            quantizer=quantizer,
        )

    @staticmethod
    def create_brevitas_conv_layer(
        in_channel, weight_quant, bias_quant, padding_type="standard"
    ):
        return bnn.QuantConv1d(
            in_channels=in_channel,
            out_channels=10,
            kernel_size=10,
            stride=1,
            padding=0,
            dilation=1,
            groups=1,
            bias=True,
            padding_type=padding_type,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
        )

    def test_depthwise_conv1d_binary_weight_bias_quant(self):
        in_channel = 3
        layer = self.create_qtorch_conv_layers(
            in_channel=in_channel, quantizer=Binarize()
        )
        target = self.create_brevitas_conv_layer(
            in_channel=in_channel,
            weight_quant=bquant.BinaryWeights,
            bias_quant=bquant.BinaryBias,
        )
        translated = translate_conv1d(layer)

        self.assertConv1dParams(target, translated)

    def test_depthwise_conv1d_ternary_weight_bias_quant(self):
        in_channel = 3
        layer = self.create_qtorch_conv_layers(
            in_channel=in_channel, quantizer=Ternarize()
        )
        target = self.create_brevitas_conv_layer(
            in_channel=in_channel,
            weight_quant=bquant.TernaryWeights,
            bias_quant=bquant.TernaryBias,
        )
        translated = translate_conv1d(layer)

        self.assertConv1dParams(target, translated)

    def test_depthwise_conv1d_padding(self):
        in_channel = 3
        layer = self.create_qtorch_conv_layers(
            in_channel=in_channel, quantizer=Binarize(), padding="same"
        )
        target = self.create_brevitas_conv_layer(
            in_channel=in_channel,
            weight_quant=bquant.BinaryWeights,
            bias_quant=bquant.BinaryBias,
            padding_type="same",
        )
        translated = translate_conv1d(layer)

        self.assertConv1dParams(target, translated)

    def test_depthwise_conv1d_wrong_padding_mode(self):
        in_channel = 3
        layer = self.create_qtorch_conv_layers(
            in_channel=in_channel, quantizer=Binarize(), padding_mode="reflect"
        )

        self.assertRaises(NotImplementedError, translate_conv1d, layer)

    def test_depthwise_conv1d_on_random_input(self):
        in_channel = 5
        layer = self.create_qtorch_conv_layers(
            in_channel=in_channel, quantizer=Ternarize()
        )

        qtorch_model = nn.Sequential(layer)

        translated_conv1d = translate_conv1d(layer)
        translated_conv1d.weight = nn.Parameter(layer.weight)
        translated_conv1d.bias = nn.Parameter(layer.bias)

        translated_model = nn.Sequential(translated_conv1d)

        random_input = torch.randn((1, 5, 1000))
        qtorch_output = qtorch_model(random_input)
        translated_output = translated_model(random_input)

        self.assertListEqual(qtorch_output.tolist(), translated_output.tolist())


if __name__ == "__main__":
    unittest.main()



import unittest

import brevitas.nn as bnn
from torch import nn

import elasticai.creator.brevitas.brevitas_quantizers as bquant
from elasticai.creator.brevitas.brevitas_model_comparison import (
    BrevitasModelComparisonTestCase,
)
from elasticai.creator.brevitas.brevitas_representation import (
    BrevitasRepresentation,
)
from elasticai.creator.systemTests.models_definition import (
    create_brevitas_model,
)


def create_small_brevitas_model():
    model = nn.Sequential(
        bnn.QuantConv1d(
            in_channels=1,
            out_channels=3,
            kernel_size=3,
            bias=False,
            weight_quant=bquant.BinaryWeights,
        ),
        nn.BatchNorm1d(num_features=3),
        bnn.QuantIdentity(act_quant=bquant.BinaryActivation),
        nn.MaxPool1d(kernel_size=5, stride=3),
        nn.Flatten(),
        bnn.QuantLinear(
            in_features=126,
            out_features=6,
            bias=False,
            weight_quant=bquant.BinaryWeights,
        ),
        nn.Sigmoid(),
    )
    return model


class SaveBrevitasModelToOnnx(BrevitasModelComparisonTestCase):
    """
    train a brevitas model and try to save it to onnx
    """

    @classmethod
    def setUpClass(cls) -> None:
        cls.brevitas_model = create_small_brevitas_model()
        cls.bigger_brevitas_model = create_brevitas_model()

    def test_save_to_onnx(self) -> None:
        br = BrevitasRepresentation(
            original_model=None, translated_model=self.brevitas_model
        )
        br.save_to_onnx((3, 1, 132), path="brevitas_model.onnx")

    def test_bigger_model_save_to_onnx(self) -> None:
        br = BrevitasRepresentation(
            original_model=None, translated_model=self.bigger_brevitas_model
        )
        br.save_to_onnx((3, 1, 97), path="bigger_brevitas_model.onnx")


if __name__ == "__main__":
    unittest.main()



from typing import Any, Dict, List, Iterable
from os import path


def get_process_string(
    component_name: str, lookup_table_generator_function: str
) -> str:
    """
    returns the string of a process Block
    Args:
        component_name (str): the name of the component
        lookup_table_generator_function (str): a function in generator_functions which generate look up table for sigmoid/tanh
    Returns:
        string as vhdl process Block
    """
    return """begin 
    {component_name}_process:process(x)
    variable int_x: integer := 0;
    begin
        int_x := to_integer(x);
        
        {generate_process}
    end process;                
    \n""".format(
        component_name=component_name, generate_process=lookup_table_generator_function
    )


def get_file_path_string(folder_names: List[str], file_name: str) -> str:
    """
    returns String of a file path
    Args:
        folder_names (List(str)): the name of the folders
        file_name (str): the name of the file
    Returns:
        string of the full path to a given filename
    """
    base_path = path.dirname(__file__)
    file_path = base_path
    for folder_name in folder_names:
        file_path = path.abspath(path.join(file_path, folder_name))
    file_path = path.abspath(path.join(file_path, file_name))
    return file_path


def get_gate_definition_string(comment: str, signal_names: List[str]) -> str:
    """
        returns the string of input gate
    Args:
        comment(str): comment for the gate definition
        signal_names (str): contains the name of signals with/without activation
    Returns:
        string of input/forget/cell/output/ gate definition or new cell state definition
    """
    signals_string = ""
    for signal in signal_names:
        signals_string += """signal {signal} : signed(DATA_WIDTH-1 downto 0):=(others=>'0');\n""".format(
            signal=signal
        )
    signals_string = comment + "\n" + signals_string + "\n\n\n"

    return signals_string


# maybe it could be more general !
def get_mac_async_architecture_behavior_string() -> str:
    """
    Returns:
        string of the behavior for mac_async architecture
    """
    return """    -- behavior: y=w1*x1+w2*x2+b
    product_1 <= shift_right((x1 * w1), FRAC_WIDTH)(DATA_WIDTH-1 downto 0);
    product_2 <= shift_right((x2 * w2), FRAC_WIDTH)(DATA_WIDTH-1 downto 0);
    y <= product_1 + product_2 + b;\n"""


def get_port_map_string(map_name: str, component_name: str, signals) -> str:
    """
    returns port map string
    Args:
        map_name (str): the name of the map
        component_name(str): the name of the component
        signals(Dict/List): dictionary with the name of each signal and its definition / List of each signal
    Returns:
        string of port map
    """
    string = """    {map_name}: {component_name}\n    port map (\n""".format(
        map_name=map_name, component_name=component_name
    )
    if type(signals) == list:
        for signal in signals:
            string += """        {signal},\n""".format(signal=signal)
    else:
        for signal in signals:
            string += """        {val1} => {val2},\n""".format(
                val1=signal, val2=signals[signal]
            )
    # remove the last comma and new line
    string = string.removesuffix(",\n") + "\n"
    # add close bracket
    string += "    );\n\n"
    return string


def get_define_process_string(
    process_name: str, sensitive_signals_list: List[str], behavior: str
) -> str:
    """
    Returns:
        string of lstm_process
    """
    signal_string = ""
    for signal in sensitive_signals_list:
        signal_string += signal + ","
    # remove the last comma
    signal_string = signal_string.removesuffix(",")

    return """    {process_name}: process({signal_string})
    begin
        {behavior}
    end process;\n\n""".format(
        process_name=process_name, signal_string=signal_string, behavior=behavior
    )



from typing import Dict

from torch.nn import Module
from elasticai.creator.layers import Binarize, Ternarize
import elasticai.creator.brevitas.brevitas_quantizers as bquant


def set_quantizers(layer: Module, args: Dict) -> None:
    """
    Function to do the mapping between brevitas and Qtorch layer quantizers.
    The function adds the quantizer to the arguments of the layer.
    Args:
        layer (Layer): current layer
        args (Dict): The layer args
    """
    if isinstance(layer.parametrizations.weight[0], Binarize):
        args["weight_quant"] = bquant.BinaryWeights
        if args["bias"]:
            args["bias_quant"] = bquant.BinaryBias
    elif isinstance(layer.parametrizations.weight[0], Ternarize):
        args["weight_quant"] = bquant.TernaryWeights
        if args["bias"]:
            args["bias_quant"] = bquant.TernaryBias



import brevitas.nn as bnn


import elasticai.creator.brevitas.brevitas_quantizers as bquant
from elasticai.creator.layers import Binarize, Ternarize


def translate_binarize_layer(layer: Binarize) -> bnn.QuantIdentity:
    """
    translates a qtorch binarize layer into a brevitas binarize layer
    Args:
        layer (Binarize): qtorch layer
    Returns:
        brevitas layer
    """
    return bnn.QuantIdentity(act_quant=bquant.BinaryActivation)


def translate_ternarize_layer(layer: Ternarize) -> bnn.QuantIdentity:
    """
    translates a qtorch ternarize layer into a brevitas ternarize layer
    Args:
        layer (Ternarize): qtorch layer
    Returns:
        brevitas layer
    """
    return bnn.QuantIdentity(act_quant=bquant.TernaryActivation)



import unittest
from os.path import exists
from elasticai.creator.vhdl.generator.functions.generate_lstm_cell_vhd import main


class GenerateLSTMCellVhdTest(unittest.TestCase):
    def setUp(self) -> None:
        main()
        self.generated_file = open("../vhdl/source/lstm_cell.vhd", "r")
        self.generated_lines = self.generated_file.readlines()
        self.expected_file = open("vhdl/vhdFiles/lstm_cell_for_testing.vhd", "r")
        self.expected_lines = self.expected_file.readlines()

    def tearDown(self) -> None:
        self.generated_file.close()
        self.expected_file.close()

    @unittest.SkipTest
    def test_generate_file(self) -> None:
        self.assertTrue(exists("../vhdl/source/lstm_cell.vhd"))

    @unittest.SkipTest
    def test_compare_files(self) -> None:
        # clean each file from empty lines and lines which are just comment
        self.expected_lines = [
            line.strip()
            for line in self.expected_lines
            if not line.startswith("--") and not line.isspace()
        ]

        self.generated_lines = [
            line.strip()
            for line in self.generated_lines
            if not line.startswith("--") and not line.isspace()
        ]

        # print("self.generated_lines= ", self.generated_lines)
        # print("self.expected_lines = ", self.expected_lines)
        self.assertEqual(self.generated_lines, self.expected_lines)
        self.assertEqual(len(self.generated_lines), len(self.expected_lines))



"""
The module contains classes and functions for generating vhdl code.
We provide code generators for the subset of vhdl that we need for implementing
our neural network accelerators and test benches. We stick closely to the vhdl
formal grammar with our class names.

The core of this module is the `CodeGenerator`. Code generators are callables that return `Code`.
`Code` is an iterable of strings. Depending on complexity we define syntactic components of the vhdl
grammar as `CodeGenerator`s. The class can then be used to set up and configure a function that yields lines
of code as strings.
"""
from collections import Sequence
from enum import Enum
from itertools import filterfalse, chain
from typing import (
    Callable,
    Iterable,
    Union,
    Literal,
    Optional,
)

Identifier = str
Code = Iterable[str]
CodeGenerator = Callable[[], Code]
CodeGeneratorCompatible = Union[Code, CodeGenerator, str]


class Keywords(Enum):
    IS = "is"
    END = "end"
    ENTITY = "entity"
    COMPONENT = "component"
    PORT = "port"
    GENERIC = "generic"
    IN = "in"
    OUT = "out"
    INOUT = "inout"
    BUFFER = "buffer"
    LINKAGE = "linkage"
    INTEGER = "integer"
    STD_LOGIC = "std_logic"


class DataType(Enum):
    INTEGER = Keywords.INTEGER.value
    STD_LOGIC = Keywords.STD_LOGIC.value


class Mode(Enum):
    IN = Keywords.IN.value
    OUT = Keywords.OUT.value
    INOUT = Keywords.INOUT.value
    BUFFER = Keywords.BUFFER.value


class _DesignUnitForEntityAndComponent:
    def __init__(
        self, identifier: str, design_type: Literal[Keywords.ENTITY, Keywords.PORT]
    ):
        self.identifier = identifier
        self._generic_list = InterfaceList()
        self._port_list = InterfaceList()
        self.type = design_type

    @property
    def generic_list(self):
        return self._generic_list

    @generic_list.setter
    def generic_list(self, value):
        self._generic_list = InterfaceList(value)

    @property
    def port_list(self):
        return self._port_list

    @port_list.setter
    def port_list(self, value):
        self._port_list = InterfaceList(value)

    def _header(self) -> Code:
        if len(self.generic_list) > 0:
            yield from _clause(Keywords.GENERIC, self._generic_list())
        if len(self.port_list) > 0:
            yield from _clause(Keywords.PORT, self._port_list())

    def __call__(self) -> Code:
        return _wrap_in_IS_END_block(self.type, self.identifier, self._header())


class Entity(_DesignUnitForEntityAndComponent):
    def __init__(self, identifier: str):
        super().__init__(identifier, Keywords.ENTITY)


class ComponentDeclaration(_DesignUnitForEntityAndComponent):
    def __init__(self, identifier: str):
        super().__init__(identifier, Keywords.COMPONENT)


class InterfaceVariable:
    def __init__(
        self,
        identifier: str,
        variable_type: DataType,
        value: Optional[Union[str, int]] = None,
        mode: Optional[Mode] = None,
    ):
        self.identifier = identifier
        self.value = value
        self.variable_type = variable_type
        self.mode = mode

    @property
    def value(self) -> int:
        return self._value

    @value.setter
    def value(self, v: Optional[Union[str, int]]):
        self._value = int(v) if v is not None else None

    def __call__(self) -> Code:
        value_part = "" if self.value is None else f" := {self.value}"
        mode_part = "" if self.mode is None else f" {self.mode} "
        yield from (
            f"{self.identifier} : {mode_part}{self.variable_type.value}{value_part}",
        )


class CodeGeneratorConcatenation(Sequence[CodeGenerator]):
    def __len__(self) -> int:
        return len(self.interface_generators)

    def __getitem__(self, item) -> CodeGenerator:
        return self.interface_generators.__getitem__(item)

    def __init__(self, *interfaces: CodeGeneratorCompatible):
        self.interface_generators = [
            _unify_code_generators(interface) for interface in interfaces
        ]

    def append(self, interface: CodeGeneratorCompatible) -> None:
        self.interface_generators.append(_unify_code_generators(interface))

    def __call__(self) -> Code:
        yield from chain.from_iterable(
            (interface() for interface in self.interface_generators)
        )


class InterfaceList(CodeGeneratorConcatenation):
    pass


InterfaceDeclaration = Union[
    "InterfaceObjectDeclaration",
    "InterfaceTypeDeclaration",
    "InterfaceSubprogramDeclaration",
    "InterfacePackageDeclaration",
]

InterfaceObjectDeclaration = Union[
    "InterfaceConstantDeclaration",
    "InterfaceSignalDeclaration",
    "InterfaceVariableDeclaration",
    "InterfaceFileDeclaration",
]

ClauseType = Literal[Keywords.GENERIC, Keywords.PORT]


def indent(line: str) -> str:
    return "".join(["\t", line])


def _add_semicolons(lines: Code) -> Code:
    temp = tuple(lines)
    yield from (f"{line};" for line in temp[:-1])
    yield f"{temp[-1]}"


def _clause(clause_type: ClauseType, interfaces: Code) -> Code:
    yield f"{clause_type.value} ("
    yield from _indent_and_filter_non_empty_lines(_add_semicolons(interfaces))
    yield ");"


def _filter_empty_lines(lines: Code) -> Code:
    return filterfalse(_line_is_empty, lines)


def _line_is_empty(line: str) -> bool:
    return len(line) == 0


def _join_lines(lines) -> str:
    return "\n".join(lines)


def _empty_code_generator() -> Code:
    return []


def _indent_and_filter_non_empty_lines(lines: Code) -> Code:
    yield from map(indent, _filter_empty_lines(lines))


# noinspection PyPep8Naming
def _wrap_in_IS_END_block(
    block_type: Keywords, block_identifier: Identifier, lines: Code
) -> Code:
    yield f"{block_type.value} {block_identifier} {Keywords.IS.value}"
    yield from _indent_and_filter_non_empty_lines(lines)
    yield f"{Keywords.END.value} {block_type.value} {block_identifier};"


def _wrap_string_into_code_generator(string: str) -> CodeGenerator:
    def wrapped():
        return (string,)

    return wrapped


def _wrap_code_into_code_generator(code: Code) -> CodeGenerator:
    def wrapped():
        return code

    return wrapped


def _unify_code_generators(generator: CodeGeneratorCompatible) -> CodeGenerator:
    if isinstance(generator, str):
        return _wrap_string_into_code_generator(generator)
    elif isinstance(generator, Iterable):
        return _wrap_code_into_code_generator(generator)
    elif isinstance(generator, Callable):
        return generator
    else:
        raise ValueError



from elasticai.creator.vhdl.generator.testbench_strings import write_testbench_file


def main(path_to_testbench: str = "../../testbench/") -> None:
    """
    generates the vhd testbench file in the testbench folder for the sigmoid function
    Args:
        path_to_testbench: path where testbench is located, default in ../testbench/
    Returns:
        None
    """
    component_name = "sigmoid"
    test_bench_file_name = component_name + "_tb.vhd"
    architecture_name = "behav"

    data_width = 16
    frac_width = 8

    components_variables_dict = {
        "x": "in signed(DATA_WIDTH-1 downto 0)",
        "y": "out signed(DATA_WIDTH-1 downto 0)",
    }
    signal_definition_dict = {
        "clk_period": "time := 1 ns",
        "test_input": "signed(16-1 downto 0):=(others=>'0')",
        "test_output": "signed(16-1 downto 0)",
    }
    uut_mapping_dict = {"x": "test_input", "y": "test_output"}

    # x, y = sigmoid(data_width, frac_width)
    # Note, the two array below, is generated based on data_width and frac_width
    # excitation signals, as test inputs signal
    inputs = [-1281, -1000, -500]
    # expected signal, as test reference output signal
    outputs = [0, 4, 28]

    write_testbench_file(
        path_to_testbench=path_to_testbench,
        test_bench_file_name=test_bench_file_name,
        component_name=component_name,
        architecture_name=architecture_name,
        data_width=data_width,
        frac_width=frac_width,
        component_variables_dict=components_variables_dict,
        signal_definitions_dict=signal_definition_dict,
        uut_mapping_dict=uut_mapping_dict,
        inputs_for_testcases=inputs,
        outputs_for_testcases=outputs,
        input_name_for_testcases="test_input",
        output_name_for_testcases="test_output",
    )


if __name__ == "__main__":
    main()



import unittest
import torch

from elasticai.creator.systemTests.models_definition import (
    create_qtorch_model,
    create_brevitas_model,
    define_weight,
)

# for reproducability
torch.manual_seed(0)


class InputSystemTest(unittest.TestCase):
    """
    System test where the same input is put in a qtorch model and the corresponding tranlsated brevitas model
    Check if the output is the same
    """

    def test_input(self) -> None:
        self.qtorch_model = create_qtorch_model()
        self.brevitas_model = create_brevitas_model()
        define_weight([layer for layer in self.qtorch_model])
        define_weight([layer for layer in self.brevitas_model])

        output_qtorch = self.qtorch_model(torch.ones(1, 1, 100))
        output_brevitas = self.brevitas_model(torch.ones(1, 1, 100))

        self.assertTrue(torch.equal(output_qtorch, output_brevitas))

    def test_bigger_input(self) -> None:
        self.qtorch_model = create_qtorch_model()
        self.brevitas_model = create_brevitas_model()
        define_weight([layer for layer in self.qtorch_model])
        define_weight([layer for layer in self.brevitas_model])

        output_qtorch = self.qtorch_model(torch.ones(512, 1, 100))
        output_brevitas = self.brevitas_model(torch.ones(512, 1, 100))

        self.assertTrue(torch.equal(output_qtorch, output_brevitas))


if __name__ == "__main__":
    unittest.main()



from typing import Any, Dict, List

from elasticai.creator.vhdl.generator.precomputed_scalar_function import (
    DataWidthVariable,
    FracWidthVariable,
)
from elasticai.creator.vhdl.generator.specific_testprocess_strings import (
    get_test_process_for_one_input_results_in_one_output_string,
    get_test_process_for_multiple_input_results_in_one_output_string,
)
from elasticai.creator.vhdl.generator.general_strings import (
    get_libraries_string,
    get_architecture_header_string,
    get_architecture_begin_string,
    get_architecture_end_string,
    get_entity_or_component_string,
    get_signal_definitions_string,
    get_variable_definitions_string,
)
from elasticai.creator.vhdl.language import (
    Entity,
    InterfaceList,
    InterfaceVariable,
    DataType,
    Mode,
)


def get_type_definitions_string(type_dict: Dict) -> str:
    """
    returns types definitions string
    Args:
        type_dict (Dict): dictionary with the name of each type and its definition
    Returns:
        string of the type definitions
    """
    type_dict_str = """    ------------------------------------------------------------
    -- Testbench Data Type
    ------------------------------------------------------------\n"""
    for type_variable in type_dict:
        type_dict_str = (
            type_dict_str
            + "    type {type_variable} is {type_definition};\n".format(
                type_variable=type_variable, type_definition=type_dict[type_variable]
            )
        )
    return type_dict_str + "\n"


def get_clock_process_string(clock_name="clk") -> str:
    """
    returns the clock process string
    Returns:
        string of the clock process
    """
    return """    clock_process : process
    begin
        {clock_name} <= '0';
        wait for clk_period/2;
        {clock_name} <= '1';
        wait for clk_period/2;
    end process; -- clock_process
\n""".format(
        clock_name=clock_name
    )


def get_uut_string(component_name: Any, mapping_dict: Dict) -> str:
    """
    writes the unit under test definition string
    Args:
        component_name (Any): name of the component
        mapping_dict (Dict): dictionary with the mapping of the variables to the signals
    Returns:
        string of the unit under test definition
    """
    mapping_dict_str = ""
    for mapping in mapping_dict:
        mapping_dict_str = (
            mapping_dict_str
            + "        "
            + mapping
            + " => "
            + mapping_dict[mapping]
            + ",\n"
        )
    # remove last comma and linebreak and add linebreak again
    return (
        """    uut: {component_name}
    port map (\n""".format(
            component_name=component_name
        )
        + mapping_dict_str[:-2]
        + "\n    );\n\n"
    )


def get_test_process_header_string() -> str:
    """
    returns test process header string
    Returns:
        string of test process header
    """
    return """    test_process: process is
    begin
        Report "======Simulation start======" severity Note;
\n"""


def get_test_process_end_string() -> str:
    """
    returns test process end string
    Returns:
        string of test process end
    """
    return """
        -- if there is no error message, that means all test case are passed.
        report "======Simulation Success======" severity Note;
        report "Please check the output message." severity Note;

        -- wait forever
        wait;

    end process; -- test_process
\n"""


def write_testbench_file(
    path_to_testbench: str,
    test_bench_file_name: str,
    component_name: str,
    architecture_name: str,
    data_width: int,
    frac_width: int,
    component_variables_dict: Dict,
    signal_definitions_dict: Dict,
    uut_mapping_dict: Dict,
    inputs_for_testcases: List,
    outputs_for_testcases: List,
    output_name_for_testcases: str,
    input_name_for_testcases: str = None,
    math_lib: bool = False,
    vector_len_width: int = None,
    type_definitions_dict: Dict = None,
    clock_name: str = "clk",
    variable_definitions_before_test_process_dict: Dict = None,
    variable_definitions_in_test_process_dict: Dict = None,
) -> None:
    """
    writes the testbench file for the specifications
    Args:
        path_to_testbench (str): path where testbench is located
        test_bench_file_name (str): name of the testbench file name
        component_name (str): component name
        architecture_name (str): architecture name
        data_width (int): data width
        frac_width (int): frac width
        component_variables_dict (Dict): variable definition of the components
        signal_definitions_dict (Dict): signal definition
        uut_mapping_dict (Dict): mapping in the uut
        inputs_for_testcases (List): List of inputs
        outputs_for_testcases (List): List of corresponding outputs for the inputs
        output_name_for_testcases (str): name of the output variable
        input_name_for_testcases (str): name of the input variable, default None (this determines if we have multiple inputs or only one)
        math_lib (bool): True, if the math library should be imported in the testbench file, default False
        vector_len_width (int): vector length width, not always defined, default None
        type_definitions_dict (Dict): possible definitions of new types, default None
        clock_name (str): name of the clock, default "clk"
        variable_definitions_before_test_process_dict (Dict): possible definitions of variables before the test process, default None
        variable_definitions_in_test_process_dict (Dict): possible definitions of variables in the test process, default None
    """
    with open(path_to_testbench + test_bench_file_name, "w") as f:
        f.write(get_libraries_string(math_lib=math_lib))
        entity = Entity(component_name + "_tb")
        entity.generic_list = [
            DataWidthVariable(value=data_width),
            FracWidthVariable(value=frac_width),
            InterfaceVariable(
                identifier="VECTOR_LEN_WIDTH", variable_type=DataType.INTEGER, value=4
            ),
        ]
        entity.port_list = [
            InterfaceVariable(
                identifier="clk", variable_type=DataType.STD_LOGIC, mode=Mode.OUT
            )
        ]
        for line in entity():
            f.write(line)
            f.write("\n")
        f.write(
            get_architecture_header_string(
                architecture_name=architecture_name,
                component_name=component_name + "_tb",
            )
        )
        if type_definitions_dict:
            f.write(get_type_definitions_string(type_dict=type_definitions_dict))
        f.write(get_signal_definitions_string(signal_dict=signal_definitions_dict))
        f.write(
            get_entity_or_component_string(
                entity_or_component="component",
                entity_or_component_name=component_name,
                data_width=data_width,
                frac_width=frac_width,
                vector_len_width=vector_len_width,
                variables_dict=component_variables_dict,
                indent="    ",
            )
        )
        f.write(get_architecture_begin_string())
        f.write(get_clock_process_string(clock_name=clock_name))
        f.write(
            get_uut_string(component_name=component_name, mapping_dict=uut_mapping_dict)
        )
        if variable_definitions_before_test_process_dict:
            f.write(
                get_variable_definitions_string(
                    variable_dict=variable_definitions_before_test_process_dict
                )
            )
        f.write(get_test_process_header_string())
        if variable_definitions_in_test_process_dict:
            f.write(
                get_variable_definitions_string(
                    variable_dict=variable_definitions_in_test_process_dict
                )
            )
        # when one input results in one output
        if input_name_for_testcases:
            f.write(
                get_test_process_for_one_input_results_in_one_output_string(
                    inputs=inputs_for_testcases,
                    outputs=outputs_for_testcases,
                    input_name=input_name_for_testcases,
                    output_name=output_name_for_testcases,
                )
            )
        else:
            f.write(
                get_test_process_for_multiple_input_results_in_one_output_string(
                    inputs=inputs_for_testcases,
                    outputs=outputs_for_testcases,
                    output_name=output_name_for_testcases,
                )
            )
        f.write(get_test_process_end_string())
        f.write(get_architecture_end_string(architecture_name=architecture_name))



import numpy as np

from elasticai.creator.vhdl.generator.precomputed_scalar_function import Tanh
from elasticai.creator.vhdl.generator.vhd_strings import (
    get_file_path_string,
)


def main():
    file_path = get_file_path_string(
        folder_names=["..", "source"], file_name="tanh.vhd"
    )
    tanh = Tanh(data_width=16, frac_width=8, x=np.linspace(-5, 5, 259))
    with open(file_path, "w") as writer:
        writer.write(tanh.build())


if __name__ == "__main__":
    main()



import unittest
from types import SimpleNamespace

from elasticai.creator.layers import Binarize, Ternarize
from elasticai.creator.brevitas.translation_functions.translation_function_tools import (
    set_quantizers,
)
import elasticai.creator.brevitas.brevitas_quantizers as bquant


class LayerMock:
    def __init__(self, quantizer):
        self.parametrizations = SimpleNamespace(weight=[quantizer])


class ConversionFunctionToolsTest(unittest.TestCase):
    def create_layers(self, quantizer):
        return LayerMock(quantizer=quantizer)

    def create_dict(self, bias):
        args = {"bias": bias}
        old_args = args.copy()
        return args, old_args

    def test_set_quantizers_binarize_quantizer_no_bias(self):
        layer = self.create_layers(quantizer=Binarize())
        args, old_args = self.create_dict(bias=False)
        set_quantizers(layer, args=args)
        self.assertNotEqual(old_args, args)
        old_args["weight_quant"] = bquant.BinaryWeights
        self.assertDictEqual(old_args, args)

    def test_set_quantizers_binarize_quantizer_with_bias(self):
        layer = self.create_layers(quantizer=Binarize())
        args, old_args = self.create_dict(bias=True)
        set_quantizers(layer, args=args)
        self.assertNotEqual(old_args, args)
        old_args["weight_quant"] = bquant.BinaryWeights
        old_args["bias_quant"] = bquant.BinaryBias
        self.assertDictEqual(old_args, args)

    def test_set_quantizers_ternarize_quantizer_no_bias(self):
        layer = self.create_layers(quantizer=Ternarize())
        args, old_args = self.create_dict(bias=False)
        set_quantizers(layer, args=args)
        self.assertNotEqual(old_args, args)
        old_args["weight_quant"] = bquant.TernaryWeights
        self.assertDictEqual(old_args, args)

    def test_set_quantizers_ternarize_quantizer_with_bias(self):
        layer = self.create_layers(quantizer=Ternarize())
        args, old_args = self.create_dict(bias=True)
        set_quantizers(layer, args=args)
        self.assertNotEqual(old_args, args)
        old_args["weight_quant"] = bquant.TernaryWeights
        old_args["bias_quant"] = bquant.TernaryBias
        self.assertDictEqual(old_args, args)


if __name__ == "__main__":
    unittest.main()



from typing import Tuple
from torch import nn

from brevitas.export import BrevitasONNXManager

from elasticai.creator.brevitas.translation_mapping import ConversionMapping
from torch.nn import Module


class BrevitasRepresentation:
    """
    This class provides an interface for the user translate from qtorch to brevitas.
    expected starting point:
    brevitas_representation = BrevitasRepresentation.from_pytorch(model)
    """

    def __init__(self, original_model: Module, translated_model: Module) -> None:
        self._original_model = original_model
        self._translated_model = translated_model

    @staticmethod
    def _translate_model(model: Module) -> Module:
        """
        translates sequential qtorch model to a sequential brevitas model
        Args:
            model (Module): qtorch model
        Returns:
            brevitas model
        """
        mapping = ConversionMapping()
        translated_layers = []

        for layer in model:
            conversion_function = mapping.get_conversion_function(layer)
            translated_layer = conversion_function(layer)
            translated_layers.append(translated_layer)

        return nn.Sequential(*translated_layers)

    @staticmethod
    def from_pytorch(model: Module):
        """
        translates sequential qtorch model to a sequential brevitas model and returns the BrevitasRepresentation
        Args:
            model (Module): qtorch model
        Returns:
            Brevitas Representation
        """
        translated_model = BrevitasRepresentation._translate_model(model)
        return BrevitasRepresentation(model, translated_model)

    @property
    def original_model(self) -> Module:
        """
        getter for original qtorch model
        Returns:
            original qtorch model
        """
        return self._original_model

    @property
    def translated_model(self) -> Module:
        """
        getter for translated brevitas model
        Returns:
            translated brevitas model
        """
        return self._translated_model

    def save_to_onnx(self, input_shape: Tuple, path: str) -> None:
        """
        saves the translated brevitas model in an onnx file with the brevitas specific conversion
        Args:
            input_shape (Tuple): input shape
            path (str): path to onnx file
        """
        # we use brevitas onnx manager because the standard onnx manager doesn't work for our model
        # (https://github.com/Xilinx/brevitas/issues/365)
        BrevitasONNXManager.export(
            module=self.translated_model, input_shape=input_shape, export_path=path
        )



from typing import Union, Tuple, List, Dict


import numpy as np


class IOTable:
    def __init__(self, inputs: np.ndarray, outputs: np.ndarray):
        self.inputs = inputs
        self.outputs = outputs
        pass

    @property
    def tables(self):
        return self.inputs, self.outputs

    def __getitem__(self, item):
        return (self.inputs[item], self.outputs[item])

    def get_table_as_dict(self) -> Dict:
        """given a list or single input,output table pair will return a list of dictionaries for each io pair. Said tables will be flatenned.
        Args:
            tables: io tables in the format
        Returns:
          dict: A list of dictionaries for the io_tables both ends will be flattened and transformed into tuples to facilitate iterating.
        """

        dic_table = {}
        n = self.tables
        for input, output in self:
            dic_table[tuple(input.flatten().tolist())] = tuple(
                output.flatten().tolist()
            )
        return dic_table


def group_tables(table: IOTable, groups: int) -> List[IOTable]:
    """given an input, output pair return a list of arrays describing the io Tables of each group.
    Args:
        table: IOTable the target IOTable
        groups: int the number of groups that will separate the said tables, those groups will break the tables along the first non-batch dimension.
    Returns:
      List[IOTable]: The grouped tables as a list of input output objects
    """
    inputs = table.tables[0]
    outputs = table.tables[1]
    assert (
        inputs.shape[1] % groups == 0 & outputs.shape[1] % groups == 0
    ), "the first dimension of the arrays should be divisible by the number of groups"
    inputs, outputs = inputs, outputs
    input_io_dim = list(inputs.shape)
    output_io_dim = list(outputs.shape)
    input_io_dim[1] = input_io_dim[1] // groups
    output_io_dim[1] = output_io_dim[1] // groups
    io_tables = [
        (np.zeros(input_io_dim), np.zeros(output_io_dim)) for _ in range(groups)
    ]
    for num, (input, output) in enumerate(zip(inputs, outputs)):
        for group in range(groups):
            input_length = input.shape[0] // groups
            output_length = output.shape[0] // groups
            io_tables[group][0][num] = input[
                group * input_length : group * input_length + input_length
            ]
            io_tables[group][1][num] = output[
                group * output_length : group * output_length + output_length
            ]
    io_tables_obj = []
    for table in io_tables:
        io_tables_obj.append(IOTable(table[0], table[1]))

    return io_tables_obj



from itertools import chain
from typing import Any, Dict

from elasticai.creator.vhdl.language import ComponentDeclaration, Entity


def get_libraries_string(math_lib: bool = False, work_lib: bool = False) -> str:
    """
    returns the string of the libraries which are imported in the vhd file
    Args:
        math_lib (bool): if True the import of the math library is added
        work_lib (bool): if True the import of the work library is added
    Returns:
        string with all library imports
    """
    lib_string = """library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;               -- for type conversions
"""
    if math_lib or work_lib:
        if math_lib:
            lib_string = lib_string + "use ieee.math_real.all;\n"
        if work_lib:
            lib_string = lib_string + "LIBRARY work;\nuse work.all;\n"
    else:
        lib_string = lib_string + "\n"
    return lib_string + "\n"


def get_entity_or_component_string(
    entity_or_component: str,
    entity_or_component_name: str,
    data_width: int,
    frac_width: int,
    variables_dict: Dict,
    vector_len_width: int = None,
    indent: str = "",
):
    """
    returns the entity or component definition string
    Args:
        entity_or_component (str):
        entity_or_component_name (str):
        data_width (int): data width
        frac_width (int): frac width
        variables_dict (Dict): dictionary with all variables and their definition
        vector_len_width (int): default not specified, if specified added to the generic part
        indent (str): number of tabs needed, specified in string
    Returns:
        string of the entity or component definition
    """
    if entity_or_component == "entity":
        entity = Entity(f"{entity_or_component_name}")

    else:
        entity = ComponentDeclaration(f"{entity_or_component_name}")

    entity.generic_list = [
        f"DATA_WIDTH : integer := {data_width}",
        f"FRAC_WIDTH : integer := {frac_width}",
    ]
    # eventually add vector_len_width
    if vector_len_width:
        entity.generic_list.append(f"VECTOR_LEN_WIDTH : integer := {vector_len_width}")

    entity.port_list = [
        f"{variable} : {definition}" for variable, definition in variables_dict.items()
    ]
    return "\n".join(chain(entity(), [""]))


def get_signal_definitions_string(signal_dict: Dict) -> str:
    """
    returns signal definitions string
    Args:
        signal_dict (Dict): dictionary with the name of each signal and its definition
    Returns:
        string of the signal definitions
    """
    signal_dict_str = ""
    for signal in signal_dict:
        signal_dict_str = (
            signal_dict_str
            + "    signal "
            + signal
            + " : "
            + signal_dict[signal]
            + ";\n"
        )
    return signal_dict_str + "\n"


def get_architecture_header_string(architecture_name: Any, component_name: Any) -> str:
    """
    returns the architecture header string
    Args:
        architecture_name (Any): name of the architecture
        component_name (Any): name of the component
    Returns:
        string with the architecture header
    """
    return """architecture {architecture_name} of {component_name} is
\n""".format(
        architecture_name=architecture_name, component_name=component_name
    )


def get_architecture_begin_string() -> str:
    """
    returns architecture begin string
    Returns:
        string of architecture begin
    """
    return "begin\n\n"


def get_architecture_end_string(architecture_name) -> str:
    """
    returns architecture end string
    Returns:
        string of architecture end
    """
    return """end architecture {architecture_name} ; -- {architecture_name}
\n""".format(
        architecture_name=architecture_name
    )


def get_variable_definitions_string(variable_dict: Dict) -> str:
    """
    returns variable definitions string in form of variable <= variable_definition
    Args:
        variable_dict (Dict): dictionary with the name of the variable and its definition
    Returns:
        string of the variable definitions
    """
    variable_str = ""
    for variable in variable_dict:
        variable_str = (
            variable_str
            + "    {variable} <= {variable_definition};\n".format(
                variable=variable, variable_definition=variable_dict[variable]
            )
        )
    return variable_str + "\n"



from elasticai.creator.vhdl.generator.general_strings import (
    get_libraries_string,
    get_entity_or_component_string,
    get_architecture_header_string,
    get_signal_definitions_string,
    get_architecture_begin_string,
    get_architecture_end_string,
)
from elasticai.creator.vhdl.generator.vhd_strings import (
    get_file_path_string,
    get_mac_async_architecture_behavior_string,
)

component_name = "mac_async"
file_name = component_name + ".vhd"
architecture_name = "mac_async_rtl"

DATA_WIDTH = 16
DATA_FRAC = 8


def main():
    file_path = get_file_path_string(folder_names=["..", "source"], file_name=file_name)

    with open(file_path, "w") as writer:
        writer.write(get_libraries_string())
        writer.write(
            get_entity_or_component_string(
                entity_or_component="entity",
                entity_or_component_name=component_name,
                data_width=DATA_WIDTH,
                frac_width=DATA_FRAC,
                variables_dict={
                    "x1": "in signed(DATA_WIDTH-1 downto 0)",
                    "x2": "in signed(DATA_WIDTH-1 downto 0)",
                    "w1": "in signed(DATA_WIDTH-1 downto 0)",
                    "w2": "in signed(DATA_WIDTH-1 downto 0)",
                    "b": "in signed(DATA_WIDTH-1 downto 0)",
                    "y": "out signed(DATA_WIDTH-1 downto 0)",
                },
            )
        )
        writer.write(
            get_architecture_header_string(
                architecture_name=architecture_name, component_name=component_name
            )
        )
        writer.write(
            get_signal_definitions_string(
                signal_dict={
                    "product_1": "signed(DATA_WIDTH-1 downto 0)",
                    "product_2": "signed(DATA_WIDTH-1 downto 0)",
                }
            )
        )
        writer.write(get_architecture_begin_string())
        writer.write(get_mac_async_architecture_behavior_string())
        writer.write(get_architecture_end_string(architecture_name=architecture_name))


if __name__ == "__main__":
    main()



from elasticai.creator.vhdl.language import (
    Entity,
    InterfaceVariable,
    DataType,
)
from unittest import TestCase


class EntityTest(TestCase):
    def test_no_name_entity(self):
        e = Entity("")
        expected = ["entity  is", "end entity ;"]
        actual = list(e())
        self.assertEqual(expected, actual)

    def test_entity_name_tanh(self):
        e = Entity("tanh")
        expected = ["entity tanh is", "end entity tanh;"]
        actual = list(e())
        self.assertEqual(expected, actual)

    def test_entity_with_generic(self):
        e = Entity("identifier")

        e.generic_list = ["test"]

        expected = [
            "entity identifier is",
            "\tgeneric (",
            "\t\ttest",
            "\t);",
            "end entity identifier;",
        ]
        actual = list(e())
        self.assertEqual(expected, actual)

    def test_entity_with_port(self):
        e = Entity("identifier")
        e.port_list = ["test"]
        expected = [
            "entity identifier is",
            "\tport (",
            "\t\ttest",
            "\t);",
            "end entity identifier;",
        ]
        actual = list(e())
        self.assertEqual(expected, actual)

    def test_entity_with_two_variables_in_generic(self):
        e = Entity("identifier")
        e.generic_list = ["first", "second"]
        expected = ["\t\tfirst;", "\t\tsecond"]
        actual = list(e())
        actual = actual[2:4]
        self.assertEqual(expected, actual)

    def test_entity_with_interface_variables(self):
        e = Entity("ident")
        e.generic_list.append(
            InterfaceVariable(
                identifier="my_var", variable_type=DataType.INTEGER, value="16"
            )
        )
        expected = ["\t\tmy_var : integer := 16"]
        actual = list(e())
        actual = actual[2:3]
        self.assertEqual(expected, actual)


example = """
entity tanh is
    generic (
        DATA_WIDTH : integer := 16;
        FRAC_WIDTH : integer := 8
    );
    port (
        x : in signed(DATA_WIDTH-1 downto 0);
        y : out signed(DATA_WIDTH-1 downto 0)
    );
end entity tanh;"""



import math
import random
import torch
import torch.nn as nn
from torch import Tensor

from elasticai.creator.layers import QLSTMCell
from typing import Dict, List, Callable, Any, Union, Sequence

from elasticai.creator.vhdl.number_representations import (
    FloatToSignedFixedPointConverter,
    two_complements_representation,
    FloatToBinaryFixedPointStringConverter,
)


def vhdl_add_assignment(code: list, line_id: str, value: str, comment=None) -> None:
    new_code_fragment = f'{line_id} <= "{value}";'
    if comment is not None:
        new_code_fragment += f" -- {comment}"
    code.append(new_code_fragment)


def precomputed_scalar_function_process(x_list, y_list) -> str:
    """
        returns the string of a lookup table
    Args:
        y_list : output List contains integers
        x_list: input List contains integers
    Returns:
        String of lookup table (if/elsif statements for vhdl file)
    """
    as_signed_fixed_point = FloatToSignedFixedPointConverter(
        bits_used_for_fraction=8, strict=False
    )
    as_binary_string = FloatToBinaryFixedPointStringConverter(
        total_bit_width=16, as_signed_fixed_point=as_signed_fixed_point
    )
    x_list.sort()
    lines = []
    if len(x_list) == 0 and len(y_list) == 1:
        vhdl_add_assignment(
            code=lines,
            line_id="y",
            value=as_binary_string(y_list[0]),
        )
    elif len(x_list) != len(y_list) - 1:
        raise ValueError(
            "x_list has to be one element shorter than y_list, but x_list has {} elements and y_list {} elements".format(
                len(x_list), len(y_list)
            )
        )
    else:
        smallest_possible_output = y_list[0]
        biggest_possible_output = y_list[-1]

        # first element
        for x in x_list[:1]:
            lines.append("if int_x<{0} then".format(as_signed_fixed_point(x)))
            vhdl_add_assignment(
                code=lines,
                line_id="y",
                value=as_binary_string(smallest_possible_output),
                comment=as_signed_fixed_point(smallest_possible_output),
            )
            lines[-1] = "\t" + lines[-1]
        for current_x, current_y in zip(x_list[1:], y_list[1:-1]):
            lines.append(
                "elsif int_x<{0} then".format(as_signed_fixed_point(current_x))
            )
            vhdl_add_assignment(
                code=lines,
                line_id="y",
                value=as_binary_string(current_y),
                comment=as_signed_fixed_point(current_y),
            )
            lines[-1] = "\t" + lines[-1]
        # last element only in y
        for y in y_list[-1:]:
            lines.append("else")
            vhdl_add_assignment(
                code=lines,
                line_id="y",
                value=as_binary_string(biggest_possible_output),
                comment=as_signed_fixed_point(biggest_possible_output),
            )
            lines[-1] = "\t" + lines[-1]
        if len(lines) != 0:
            lines.append("end if;")
    # build the string block and add new line + 2 tabs
    string = ""
    for line in lines:
        string = string + line + "\n" + "\t" + "\t"
    return string


def _int_to_hex(val: int, nbits: int) -> str:
    """
        returns a string of a hexadecimal value for an integer
    Args:
        val (int): the integer value which will converted in hexadecimal
        nbits(int): the number of bits
    Returns:
        String of a hexadecimal value for an integer
    """
    if val < 0:
        return hex((val + (1 << nbits)) % (1 << nbits))
    else:
        return "{0:#0{1}x}".format(val, 2 + int(nbits / 4))


def _floating_to_hex(f_val: float, frac_width: int, nbits: int) -> str:
    """
        returns a string of a hexadecimal value for a floating point number
    Args:
        f_val (int): the floating point number which will converted to hexadecimal
        nbits(int): the number of bits
        frac_width(int): the number of bits for the fraction
    Returns:
        String of a hexadecimal value for a floating number
    """
    int_val = int(f_val * (2 ** frac_width))
    return _int_to_hex(int_val, nbits)


def _to_vhdl_parameter(
    f_val: float, frac_width: int, nbits: int, name_parameter: str, signal_name: str
) -> Dict:
    """
        returns a Dictionary of one signal and his definition
    Args:
        f_val (float): a floating point number
        nbits(int): the number of bits
        frac_width(int): the number of bits for the fraction
    Returns:
        String of a hexadecimal value for an integer
    """
    hex_str = _floating_to_hex(f_val, frac_width, nbits)
    hex_str_without_prefix = hex_str[2:]

    return {
        str(signal_name): "signed(DATA_WIDTH-1 downto 0) := "
        + 'X"'
        + hex_str_without_prefix
        + '"; -- '
        + name_parameter
    }


def _elastic_ai_creator_lstm() -> QLSTMCell:

    return QLSTMCell(
        1, 1, state_quantizer=nn.Identity(), weight_quantizer=nn.Identity()
    )


def _ensure_reproducibility():
    torch.manual_seed(0)
    random.seed(0)


def generate_signal_definitions_for_lstm(data_width: int, frac_width: int) -> Dict:
    """
    returns Dict of signals names as key and their definition as value
    Args:
        data_width (int): the width of the data
        frac_width (int): the fraction part of data_width
    Returns:
        Dict of the signal names and their definitions
    """
    dict_of_signals = {}
    # define the lstm cell
    _ensure_reproducibility()
    lstm_single_cell = _elastic_ai_creator_lstm()

    # weight_ih_l[k] : `(W_ii|W_if|W_ig|W_io)`
    # weight_hh_l[k] : `(W_hi|W_hf|W_hg|W_ho)`
    # bias_ih_l[k] :  `(b_ii|b_if|b_ig|b_io)`
    # bias_hh_l[k] :  `(W_hi|W_hf|W_hg|b_ho)`
    b_ii = 0
    b_if = 0
    b_ig = 0
    b_io = 0

    b_hi = 0
    b_hf = 0
    b_hg = 0
    b_ho = 0

    for name, param in lstm_single_cell.named_parameters():
        if name == "weight_ih":
            dict_of_signals.update(
                _to_vhdl_parameter(
                    param[0],
                    frac_width,
                    data_width,
                    name_parameter="W_ii",
                    signal_name="wii",
                )
            )
            dict_of_signals.update(
                _to_vhdl_parameter(
                    param[1],
                    frac_width,
                    data_width,
                    name_parameter="W_if",
                    signal_name="wif",
                )
            )
            dict_of_signals.update(
                _to_vhdl_parameter(
                    param[2],
                    frac_width,
                    data_width,
                    name_parameter="W_ig",
                    signal_name="wig",
                )
            )
            dict_of_signals.update(
                _to_vhdl_parameter(
                    param[3],
                    frac_width,
                    data_width,
                    name_parameter="W_io",
                    signal_name="wio",
                )
            )
        elif name == "weight_hh":
            dict_of_signals.update(
                _to_vhdl_parameter(
                    param[0],
                    frac_width,
                    data_width,
                    name_parameter="W_hi",
                    signal_name="whi",
                )
            )
            dict_of_signals.update(
                _to_vhdl_parameter(
                    param[1],
                    frac_width,
                    data_width,
                    name_parameter="W_hf",
                    signal_name="whf",
                )
            )
            dict_of_signals.update(
                _to_vhdl_parameter(
                    param[2],
                    frac_width,
                    data_width,
                    name_parameter="W_hg",
                    signal_name="whg",
                )
            )
            dict_of_signals.update(
                _to_vhdl_parameter(
                    param[3],
                    frac_width,
                    data_width,
                    name_parameter="W_ho",
                    signal_name="who",
                )
            )
        elif name == "bias_ih":
            b_ii = param[0]
            b_if = param[1]
            b_ig = param[2]
            b_io = param[3]
        elif name == "bias_hh":
            b_hi = param[0]
            b_hf = param[1]
            b_hg = param[2]
            b_ho = param[3]
        else:
            dict_of_signals.update("should not come to here.")

    dict_of_signals.update(
        _to_vhdl_parameter(
            b_ii + b_hi,
            frac_width,
            data_width,
            name_parameter="b_ii + b_hi",
            signal_name="bi",
        )
    )
    dict_of_signals.update(
        _to_vhdl_parameter(
            b_if + b_hf,
            frac_width,
            data_width,
            name_parameter="b_if + b_hf",
            signal_name="bf",
        )
    )
    dict_of_signals.update(
        _to_vhdl_parameter(
            b_ig + b_hg,
            frac_width,
            data_width,
            name_parameter="b_ig + b_hg",
            signal_name="bg",
        )
    )
    dict_of_signals.update(
        _to_vhdl_parameter(
            b_io + b_ho,
            frac_width,
            data_width,
            name_parameter="b_io + b_ho",
            signal_name="bo",
        )
    )

    return dict_of_signals



import unittest
import torch
from torch import nn

import brevitas.nn as bnn

from elasticai.creator.brevitas.translation_functions.conv import (
    translate_conv2d,
)
from elasticai.creator.integrationTests.conv_params_comparison import (
    ConvTest,
)
from elasticai.creator.layers import QConv2d, Binarize, Ternarize
import elasticai.creator.brevitas.brevitas_quantizers as bquant


class Conv2dTest(ConvTest):
    @staticmethod
    def create_qtorch_conv_layers(quantizer, padding=0, padding_mode="zeros"):
        return QConv2d(
            in_channels=5,
            out_channels=10,
            kernel_size=10,
            stride=1,
            padding=padding,
            dilation=1,
            groups=1,
            bias=True,
            padding_mode=padding_mode,
            quantizer=quantizer,
        )

    @staticmethod
    def create_brevitas_conv_layer(weight_quant, bias_quant, padding_type="standard"):
        return bnn.QuantConv2d(
            in_channels=5,
            out_channels=10,
            kernel_size=10,
            stride=1,
            padding=0,
            dilation=1,
            groups=1,
            bias=True,
            padding_type=padding_type,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
        )

    def test_conv2d_binary_weight_bias_quant(self):
        layer = self.create_qtorch_conv_layers(quantizer=Binarize())
        target = self.create_brevitas_conv_layer(
            weight_quant=bquant.BinaryWeights, bias_quant=bquant.BinaryBias
        )
        translated = translate_conv2d(layer)

        self.assertConv2dParams(target, translated)

    def test_conv2d_ternaryweight_bias_quant(self):
        layer = self.create_qtorch_conv_layers(quantizer=Ternarize())
        target = self.create_brevitas_conv_layer(
            weight_quant=bquant.TernaryWeights, bias_quant=bquant.TernaryBias
        )
        translated = translate_conv2d(layer)

        self.assertConv2dParams(target, translated)

    def test_conv2d_padding(self):
        layer = self.create_qtorch_conv_layers(quantizer=Binarize(), padding="same")
        target = self.create_brevitas_conv_layer(
            weight_quant=bquant.BinaryWeights,
            bias_quant=bquant.BinaryBias,
            padding_type="same",
        )
        translated = translate_conv2d(layer)

        self.assertConv2dParams(target, translated)

    def test_conv2d_wrong_padding_mode(self):
        layer = self.create_qtorch_conv_layers(
            quantizer=Binarize(), padding_mode="reflect"
        )
        self.assertRaises(NotImplementedError, translate_conv2d, layer)

    def test_conv2d_on_random_input(self):
        layer = self.create_qtorch_conv_layers(quantizer=Ternarize())
        qtorch_model = nn.Sequential(layer)

        translated_conv2d = translate_conv2d(layer)
        translated_conv2d.weight = nn.Parameter(layer.weight)
        translated_conv2d.bias = nn.Parameter(layer.bias)

        translated_model = nn.Sequential(translated_conv2d)

        random_input = torch.randn((10, 5, 10, 10))
        qtorch_output = qtorch_model(random_input)
        translated_output = translated_model(random_input)

        self.assertListEqual(qtorch_output.tolist(), translated_output.tolist())


if __name__ == "__main__":
    unittest.main()



import torch
from torch import nn

import brevitas.nn as bnn
from torch.nn.utils.parametrize import is_parametrized

from elasticai.creator.layers import Binarize, QConv1d, QLinear
import elasticai.creator.brevitas.brevitas_quantizers as bquant


def create_qtorch_model():
    model = nn.Sequential(
        QConv1d(
            in_channels=1,
            out_channels=3,
            kernel_size=3,
            quantizer=Binarize(),
            bias=False,
        ),
        nn.BatchNorm1d(num_features=3),
        Binarize(),
        nn.MaxPool1d(
            kernel_size=5,
            stride=3,
        ),
        QConv1d(
            in_channels=3,
            out_channels=3,
            kernel_size=3,
            quantizer=Binarize(),
        ),
        nn.BatchNorm1d(num_features=3),
        Binarize(),
        QConv1d(
            in_channels=3,
            out_channels=3,
            kernel_size=3,
            quantizer=Binarize(),
        ),
        nn.BatchNorm1d(num_features=3),
        Binarize(),
        nn.MaxPool1d(
            kernel_size=3,
            stride=2,
        ),
        QConv1d(
            in_channels=3,
            out_channels=3,
            kernel_size=2,
            quantizer=Binarize(),
        ),
        nn.BatchNorm1d(num_features=3),
        Binarize(),
        QConv1d(
            in_channels=3,
            out_channels=3,
            kernel_size=2,
            quantizer=Binarize(),
        ),
        nn.BatchNorm1d(num_features=3),
        Binarize(),
        nn.MaxPool1d(
            kernel_size=3,
            stride=2,
        ),
        nn.Flatten(),
        QLinear(in_features=15, out_features=1, quantizer=Binarize(), bias=False),
        nn.Sigmoid(),
    )
    return model


def create_brevitas_model():
    model = nn.Sequential(
        bnn.QuantConv1d(
            in_channels=1,
            out_channels=3,
            kernel_size=3,
            bias=False,
            weight_quant=bquant.BinaryWeights,
        ),
        nn.BatchNorm1d(num_features=3),
        bnn.QuantIdentity(act_quant=bquant.BinaryActivation),
        nn.MaxPool1d(kernel_size=5, stride=3),
        bnn.QuantConv1d(
            in_channels=3,
            out_channels=3,
            kernel_size=3,
            weight_quant=bquant.BinaryWeights,
            bias_quant=bquant.BinaryBias,
        ),
        nn.BatchNorm1d(num_features=3),
        bnn.QuantIdentity(act_quant=bquant.BinaryActivation),
        bnn.QuantConv1d(
            in_channels=3,
            out_channels=3,
            kernel_size=3,
            weight_quant=bquant.BinaryWeights,
            bias_quant=bquant.BinaryBias,
        ),
        nn.BatchNorm1d(num_features=3),
        bnn.QuantIdentity(act_quant=bquant.BinaryActivation),
        nn.MaxPool1d(kernel_size=3, stride=2),
        bnn.QuantConv1d(
            in_channels=3,
            out_channels=3,
            kernel_size=2,
            weight_quant=bquant.BinaryWeights,
            bias_quant=bquant.BinaryBias,
        ),
        nn.BatchNorm1d(num_features=3),
        bnn.QuantIdentity(act_quant=bquant.BinaryActivation),
        bnn.QuantConv1d(
            in_channels=3,
            out_channels=3,
            kernel_size=2,
            weight_quant=bquant.BinaryWeights,
            bias_quant=bquant.BinaryBias,
        ),
        nn.BatchNorm1d(num_features=3),
        bnn.QuantIdentity(act_quant=bquant.BinaryActivation),
        nn.MaxPool1d(
            kernel_size=3,
            stride=2,
        ),
        nn.Flatten(),
        bnn.QuantLinear(
            in_features=15,
            out_features=1,
            bias=False,
            weight_quant=bquant.BinaryWeights,
        ),
        nn.Sigmoid(),
    )
    return model


weight_layers = (QConv1d, QLinear, nn.BatchNorm1d, bnn.QuantConv1d, bnn.QuantLinear)


def define_weight(layers):
    # TODO: In case we're using this function just to have deterministic weight value generation
    #  we should reconsider our approach here
    for layer in layers:
        if isinstance(layer, weight_layers):
            if is_parametrized(layer):
                layer.weight = torch.ones_like(layer.weight)
            else:
                layer.weight = torch.nn.Parameter(torch.ones_like(layer.weight))
            if layer.bias is not None:
                if is_parametrized(layer):
                    layer.bias = torch.ones_like(layer.bias)
                else:
                    layer.bias = torch.nn.Parameter(torch.ones_like(layer.bias))



from torch import nn

import brevitas.nn as bnn

from elasticai.creator.brevitas.translation_functions.quantizers import (
    translate_binarize_layer,
    translate_ternarize_layer,
)
from elasticai.creator.brevitas.brevitas_model_comparison import (
    BrevitasModelComparisonTestCase,
)
import elasticai.creator.brevitas.brevitas_quantizers as bquant
from elasticai.creator.layers import Binarize, Ternarize


class QuantizersTest(BrevitasModelComparisonTestCase):
    def test_binarize(self):
        layer = Binarize()
        target = nn.Sequential(bnn.QuantIdentity(act_quant=bquant.BinaryActivation))
        translated_layers = translate_binarize_layer(layer)
        translated_model = nn.Sequential(translated_layers)

        self.assertModelEqual(target, translated_model)

    def test_ternarize(self):
        layer = Ternarize()
        target = nn.Sequential(bnn.QuantIdentity(act_quant=bquant.TernaryActivation))
        translated_layers = translate_ternarize_layer(layer)
        translated_model = nn.Sequential(translated_layers)

        self.assertModelEqual(target, translated_model)



import math
from abc import ABC, abstractmethod
from itertools import chain

import torch.nn

from elasticai.creator.vhdl.generator.general_strings import (
    get_libraries_string,
    get_entity_or_component_string,
    get_architecture_header_string,
    get_architecture_end_string,
)
from elasticai.creator.vhdl.generator.generator_functions import (
    precomputed_scalar_function_process,
)
from elasticai.creator.vhdl.generator.vhd_strings import get_process_string
from elasticai.creator.vhdl.language import (
    Entity,
    InterfaceVariable,
    DataType,
)


class DataWidthVariable(InterfaceVariable):
    def __init__(self, value: int):
        super().__init__(
            identifier="DATA_WIDTH", variable_type=DataType.INTEGER, value=f"{value}"
        )


class FracWidthVariable(InterfaceVariable):
    def __init__(self, value: int):
        super().__init__(
            identifier="FRAC_WIDTH", variable_type=DataType.INTEGER, value=f"{value}"
        )


class PrecomputedScalarFunction:
    def __init__(
        self, data_width, frac_width, x, y, component_name=None, process_instance=None
    ):
        self.component_name = self._get_lower_case_class_name_or_component_name(
            component_name=component_name
        )
        self.data_width = data_width
        self.frac_width = frac_width
        self.x = x
        self.y = y
        self.process_instance = process_instance

    @classmethod
    def _get_lower_case_class_name_or_component_name(cls, component_name):
        if component_name is None:
            return cls.__name__.lower()
        return component_name

    @property
    def file_name(self) -> str:
        return f"{self.component_name}.vhd"

    @property
    def architecture_name(self) -> str:
        return f"{self.component_name}_rtl"

    def __call__(self) -> list[str]:
        entity = Entity(self.component_name)
        entity.generic_list = [
            f"DATA_WIDTH : integer := {self.data_width}",
            f"FRAC_WIDTH : integer := {self.frac_width}",
        ]
        entity.port_list = [
            "x : in signed(DATA_WIDTH-1 downto 0)",
            "y : out signed(DATA_WIDTH-1 downto 0)",
        ]
        code = "\n".join(chain([get_libraries_string()], chain(entity()), [""]))
        code += get_architecture_header_string(
            architecture_name=self.architecture_name, component_name=self.component_name
        )
        code += get_process_string(
            component_name=self.component_name,
            lookup_table_generator_function=precomputed_scalar_function_process(
                x_list=self.x, y_list=self.y
            ),
        )
        code += get_architecture_end_string(architecture_name=self.architecture_name)
        return [code]


class Sigmoid(PrecomputedScalarFunction):
    def __init__(self, data_width, frac_width, x, component_name=None):
        x_list = torch.as_tensor(x)
        # calculate y always for the previous element, therefore the last input is not needed here
        y_list = list(torch.nn.Sigmoid()(x_list[:-1]))
        y_list.insert(0, 0)
        # add last y value, therefore, x_list is one element shorter than y_list
        y_list.append(1)
        super(Sigmoid, self).__init__(
            data_width=data_width,
            frac_width=frac_width,
            x=x,
            y=y_list,
            component_name=component_name,
        )

    def build(self) -> str:
        lines_of_code = self.__call__()
        return "\n".join(lines_of_code)


class Tanh(PrecomputedScalarFunction):
    def __init__(self, data_width, frac_width, x, component_name=None):
        y_list = [-1]
        # calculate y always for the previous element, therefore the last input is not needed here
        for x_element in x[:-1]:
            y_list.append(math.tanh(x_element))
        # add last y value, therefore, x_list is one element shorter than y_list
        y_list.append(1)
        super(Tanh, self).__init__(
            data_width=data_width,
            frac_width=frac_width,
            x=x,
            y=y_list,
            component_name=component_name,
        )

    def build(self) -> str:
        lines_of_code = self.__call__()
        return "\n".join(lines_of_code)



"""
make conversion function of each layer usable in other packages
"""
from elasticai.creator.brevitas.translation_functions.linear import (
    translate_linear_layer,
)
from elasticai.creator.brevitas.translation_functions.identity import (
    translate_layer,
)
from elasticai.creator.brevitas.translation_functions.conv import (
    translate_conv1d,
    translate_conv2d,
)
from elasticai.creator.brevitas.translation_functions.quantizers import (
    translate_binarize_layer,
    translate_ternarize_layer,
)



from unittest import TestCase
from elasticai.creator.vhdl.number_representations import (
    FloatToSignedFixedPointConverter,
    two_complements_representation,
)


class FixedPointConverterTest(TestCase):
    def test_get_zero(self):
        f = FloatToSignedFixedPointConverter(bits_used_for_fraction=0)
        self.assertEqual(0, f(0))

    def test_get_one_with_2bits_for_fraction(self):
        f = FloatToSignedFixedPointConverter(bits_used_for_fraction=2)
        self.assertEqual(1 << 2, f(1))

    def test_get_one_with_3bits_for_fraction(self):
        f = FloatToSignedFixedPointConverter(bits_used_for_fraction=3)
        self.assertEqual(1 << 3, f(1))

    def test_raise_error_if_not_convertible(self):
        f = FloatToSignedFixedPointConverter(bits_used_for_fraction=0)
        try:
            f(0.5)
            self.fail()
        except ValueError as e:
            self.assertEqual(
                "0.5 not convertible to fixed point number using 0 bits for fractional part",
                str(e),
            )


class BinaryTwoComplementRepresentation(TestCase):
    def test_one(self):
        actual = two_complements_representation(1, 1)
        expected = "1"
        self.assertEqual(expected, actual)

    def test_minus_one(self):
        actual = two_complements_representation(-1, 2)
        expected = "11"
        self.assertEqual(expected, actual)

    def test_minus_two(self):
        actual = two_complements_representation(-2, 2)
        expected = "10"
        self.assertEqual(expected, actual)

    def test_two(self):
        actual = two_complements_representation(2, 3)
        expected = "010"
        self.assertEqual(expected, actual)

    def test_minus_four(self):
        actual = two_complements_representation(-4, 3)
        expected = "100"
        self.assertEqual(expected, actual)

    def test_minus_three_three_bit(self):
        actual = two_complements_representation(-3, 3)
        expected = "101"
        self.assertEqual(expected, actual)

    def test_minus_256_16_bit(self):
        actual = two_complements_representation(-256, 16)
        expected = "1111111100000000"
        self.assertEqual(expected, actual)

    def test_minus_254_16_bit(self):
        actual = two_complements_representation(-254, 16)
        expected = "1111111100000010"
        self.assertEqual(expected, actual)



from elasticai.creator.vhdl.generator.testbench_strings import write_testbench_file


def main(path_to_testbench: str = "../../testbench/") -> None:
    """
    generates the vhd testbench file in the testbench folder for lstm common gate
    Args:
        path_to_testbench: path where testbench is located, default in ../testbench/
    Returns:
        None
    """
    component_name = "lstm_common_gate"
    test_bench_file_name = component_name + "_tb.vhd"
    architecture_name = "arch"

    data_width = 16
    frac_width = 8
    vector_len_width = 4

    components_variables_dict = {
        "reset": "in std_logic",
        "clk": "in std_logic",
        "x": "in signed(DATA_WIDTH-1 downto 0)",
        "w": "in signed(DATA_WIDTH-1 downto 0)",
        "b": "in signed(DATA_WIDTH-1 downto 0)",
        "vector_len": "in unsigned(VECTOR_LEN_WIDTH-1 downto 0)",
        "idx": "out unsigned(VECTOR_LEN_WIDTH-1 downto 0)",
        "ready": "out std_logic",
        "y": "out signed(DATA_WIDTH-1 downto 0)",
    }
    type_definitions_dict = {
        "RAM_ARRAY": "array (0 to 9 ) of signed(DATA_WIDTH-1 downto 0)"
    }
    signal_definition_dict = {
        "clk_period": "time := 2 ps",
        "clock": "std_logic",
        "reset, ready": "std_logic:='0'",
        "X_MEM": "RAM_ARRAY :=(others=>(others=>'0'))",
        "W_MEM": "RAM_ARRAY:=(others=>(others=>'0'))",
        "x, w, y, b": "signed(DATA_WIDTH-1 downto 0):=(others=>'0')",
        "vector_len": "unsigned(VECTOR_LEN_WIDTH-1 downto 0):=(others=>'0')",
        "idx": "unsigned(VECTOR_LEN_WIDTH-1 downto 0):=(others=>'0')",
    }
    uut_mapping_dict = {
        "reset": "reset",
        "clk": "clock",
        "x": "x",
        "w": "w",
        "b": "b",
        "vector_len": "vector_len",
        "idx": "idx",
        "ready": "ready",
        "y": "y",
    }
    variable_definitions_before_test_process_dict = {
        "x": "X_MEM(to_integer(idx))",
        "w": "W_MEM(to_integer(idx))",
    }
    variable_definitions_in_test_process_dict = {
        "    vector_len": "to_unsigned(10, VECTOR_LEN_WIDTH)"
    }
    # Note, the two array below, is generated based on data_width and frac_width
    # excitation signals, as test inputs signal
    inputs = [
        {
            "X_MEM": '(x"0013",x"0000",x"0010",x"0013",x"000c",x"0005",x"0005",x"0013",x"0004",x"0002")',
            "W_MEM": '(x"0011",x"0018",x"0000",x"000d",x"0014",x"000f",x"0012",x"0007",x"0017",x"0012")',
            "b": 'x"008a"',
        },
        {
            "X_MEM": '(x"0014",x"000d",x"0017",x"0008",x"0002",x"0007",x"0002",x"0015",x"0001",x"0010")',
            "W_MEM": '(x"000e",x"0014",x"0005",x"0015",x"0009",x"0013",x"0007",x"0016",x"0008",x"0004")',
            "b": 'x"0064"',
        },
        {
            "X_MEM": '(x"000f",x"0017",x"000d",x"000f",x"0001",x"0009",x"0002",x"0007",x"0008",x"0013")',
            "W_MEM": '(x"0001",x"000a",x"0008",x"0010",x"0008",x"0001",x"0016",x"0013",x"0016",x"000a")',
            "b": 'x"009b"',
        },
        {
            "X_MEM": '(x"000c",x"0007",x"0001",x"0019",x"0008",x"000c",x"0019",x"000b",x"0008",x"000d")',
            "W_MEM": '(x"000e",x"0015",x"0001",x"000b",x"0014",x"0012",x"000f",x"0000",x"0008",x"000e")',
            "b": 'x"004c"',
        },
        {
            "X_MEM": '(x"0005",x"0013",x"0002",x"0013",x"000c",x"000f",x"0003",x"0004",x"0010",x"0001")',
            "W_MEM": '(x"0006",x"000d",x"0005",x"0009",x"0017",x"0017",x"000e",x"000d",x"0000",x"0019")',
            "b": 'x"0092"',
        },
    ]
    # expected signal, as test reference output signal
    outputs = [142, 105, 159, 82, 150]

    write_testbench_file(
        path_to_testbench=path_to_testbench,
        test_bench_file_name=test_bench_file_name,
        component_name=component_name,
        architecture_name=architecture_name,
        data_width=data_width,
        frac_width=frac_width,
        component_variables_dict=components_variables_dict,
        signal_definitions_dict=signal_definition_dict,
        uut_mapping_dict=uut_mapping_dict,
        inputs_for_testcases=inputs,
        outputs_for_testcases=outputs,
        output_name_for_testcases="y",
        math_lib=True,
        vector_len_width=vector_len_width,
        type_definitions_dict=type_definitions_dict,
        clock_name="clock",
        variable_definitions_before_test_process_dict=variable_definitions_before_test_process_dict,
        variable_definitions_in_test_process_dict=variable_definitions_in_test_process_dict,
    )


if __name__ == "__main__":
    main()



import unittest


class GeneratedVHDLCodeTest(unittest.TestCase):
    @staticmethod
    def get_non_empty_lines(text):
        text = text.splitlines()
        return [x for x in text if len(x) > 0]

    def check_generated_code(self, expected_code, code):
        expected = self.get_non_empty_lines(expected_code)
        actual = self.get_non_empty_lines(code)
        expected = [
            line.strip()
            for line in expected
            if not line.startswith("--") and not line.isspace()
        ]

        actual = [
            line.strip()
            for line in actual
            if not line.startswith("--") and not line.isspace()
        ]
        self.assertEqual(expected, actual)

    expected_code = None



from typing import Union, overload


class FloatToSignedFixedPointConverter:
    """
    Create a fixed point representation as an unsigned int data type using two complements.

    We might want to have this create its own type `FixedPointNumber` in
    the future. That way we could make sure that the conversion is idempotent
    for numbers that are fixed point already.
    """

    def __init__(self, bits_used_for_fraction: int, strict=True):
        self.bits_used_for_fraction = bits_used_for_fraction
        self._strict = strict

    @property
    def one(self) -> int:
        return 1 << self.bits_used_for_fraction

    def __call__(self, x: float) -> int:
        x_tmp = float(x)
        x_tmp = x_tmp * self.one
        if not x_tmp.is_integer() and self._strict:
            raise ValueError(
                f"{x} not convertible to fixed point number using {self.bits_used_for_fraction} bits for fractional part"
            )
        return int(x_tmp)

    def to_string(self, x: float) -> str:
        return str(self.__call__(x))


class FloatToBinaryFixedPointStringConverter:
    def __init__(
        self,
        total_bit_width: int,
        as_signed_fixed_point: FloatToSignedFixedPointConverter,
    ):
        self.total_bit_width = total_bit_width
        self.as_signed_fixed_point = as_signed_fixed_point

    def __call__(self, x: Union[float, int]) -> str:
        signed_fixed_point = self.as_signed_fixed_point(x)
        return two_complements_representation(signed_fixed_point, self.total_bit_width)


def two_complements_representation(x, number_of_bits):
    if x < 0:
        unsigned_int_version = (1 << number_of_bits) + x
    else:
        unsigned_int_version = x
    return "{{0:0{number_of_bits}b}}".format(number_of_bits=number_of_bits).format(
        unsigned_int_version
    )



import unittest

import brevitas.nn as bnn


class ConvTest(unittest.TestCase):
    """
    self implemented Test case for comparing two brevitas convolutional layers
    """

    def assertConvParams(self, target, translated):
        self.assertEqual(target.in_channels, translated.in_channels)
        self.assertEqual(target.out_channels, translated.out_channels)
        self.assertEqual(target.kernel_size, translated.kernel_size)
        self.assertEqual(target.stride, translated.stride)
        self.assertEqual(target.padding, translated.padding)
        self.assertEqual(target.dilation, translated.dilation)
        self.assertEqual(target.groups, translated.groups)
        self.assertEqual(target.bias is not None, translated.bias is not None)
        self.assertEqual(target.padding_type, translated.padding_type)
        self.assertEqual(str(target.weight_quant), str(translated.weight_quant))
        self.assertEqual(str(target.bias_quant), str(translated.bias_quant))
        self.assertEqual(str(target), str(translated))

    def assertConv1dParams(self, target, translated):
        self.assertIsInstance(translated, bnn.QuantConv1d)
        self.assertConvParams(target, translated)

    def assertConv2dParams(self, target, translated):
        self.assertIsInstance(translated, bnn.QuantConv2d)
        self.assertConvParams(target, translated)



import unittest

from elasticai.creator.brevitas.brevitas_model_comparison import (
    BrevitasModelComparisonTestCase,
)
from elasticai.creator.brevitas.brevitas_representation import (
    BrevitasRepresentation,
)
from elasticai.creator.systemTests.models_definition import (
    create_qtorch_model,
    create_brevitas_model,
    define_weight,
)


class ModelSystemTest(BrevitasModelComparisonTestCase):
    """
    System tests for translating a big qtorch model to brevitas
    """

    def test_complete_models_with_weights(self) -> None:
        self.qtorch_model = create_qtorch_model()

        self.brevitas_model = create_brevitas_model()
        define_weight([layer for layer in self.brevitas_model])

        translated_model = BrevitasRepresentation.from_pytorch(
            self.qtorch_model
        ).translated_model
        translated_layers = [layer for layer in translated_model]
        define_weight(translated_layers)

        self.assertModelEqual(translated_model, self.brevitas_model)


if __name__ == "__main__":
    unittest.main()



from typing import List, Any, Dict


def get_test_process_for_one_input_results_in_one_output_string(
    inputs: List[Any], outputs: List[Any], input_name: str, output_name: str
) -> str:
    """
    writes test process cases for a function like sigmoid or tanh
    Args:
        inputs (List[Any]): list of the inputs
        outputs (List[Outputs]): list of outputs
    Returns:
        string of the testcases of each input and output pair
    """

    test = ""
    if len(inputs) == len(outputs):
        for i in range(len(inputs)):
            test = test + """        {input_name} <=  to_signed({input},16);
        wait for 1*clk_period;
        report "The value of '{output_name}' is " & integer'image(to_integer(unsigned({output_name})));
        assert {output_name}={output} report "The test case {input} fail" severity failure;
\n""".format(
                input=inputs[i],
                output=outputs[i],
                input_name=input_name,
                output_name=output_name,
            )
        return test
    else:
        raise TypeError(
            f"inputs length {len(inputs)} is different to outputs length {len(outputs)}."
        )


def get_test_process_for_multiple_input_results_in_one_output_string(
    inputs: List[Dict], outputs: List[Any], output_name: str
) -> str:
    """
    returns test process cases for multiple inputs which result in one output
    Args:
        inputs (List[Dict]): list of inputs for one output, the list includes a dictionary which defines each input
        outputs (List): output for each test case
        output_name (str): name of the output variable
    Returns:
        string of the testcases of each input dictionary and output pair
    """
    test = ""
    if len(inputs) == len(outputs):
        for i in range(len(outputs)):
            input_dict = inputs[i]
            for input in input_dict:
                test = test + "        {input_name} <= {input};\n".format(
                    input_name=input, input=input_dict[input]
                )
            test = (
                test
                + """
        reset <= '1';
        wait for 2*clk_period;
        wait until clock = '0';
        reset <= '0';
        wait until ready = '1';

        report "expected output is {output}, value of '{output_name}' is " & integer'image(to_integer(signed({output_name})));
        assert {output_name} = {output} report "The {counter}. test case fail" severity error;
        reset <= '1';
        wait for 1*clk_period;
\n""".format(
                    output=outputs[i], output_name=output_name, counter=i
                )
            )
        return test
    else:
        raise TypeError(
            f"inputs length {len(inputs)} is different to outputs length {len(outputs)}."
        )



import unittest
import torch
from torch import nn

import brevitas.nn as bnn

from elasticai.creator.brevitas.translation_functions.linear import (
    translate_linear_layer,
)
from elasticai.creator.layers import QLinear, Binarize, Ternarize
import elasticai.creator.brevitas.brevitas_quantizers as bquant


class LinearLayerTest(unittest.TestCase):
    @staticmethod
    def create_qtorch_linear_layer(quantizer):
        return QLinear(in_features=5, out_features=10, bias=True, quantizer=quantizer)

    @staticmethod
    def create_brevitas_linear_layer(weight_quant, bias_quant):
        return bnn.QuantLinear(
            in_features=5,
            out_features=10,
            bias=True,
            weight_quant=weight_quant,
            bias_quant=bias_quant,
        )

    def test_translate_linear_layer_binary_weight_bias_quant(self):
        layer = self.create_qtorch_linear_layer(quantizer=Binarize())
        target = self.create_brevitas_linear_layer(
            weight_quant=bquant.BinaryWeights, bias_quant=bquant.BinaryBias
        )
        translated = translate_linear_layer(layer)

        self.assertIsInstance(translated, bnn.QuantLinear)
        self.assertEqual(target.in_features, translated.in_features)
        self.assertEqual(target.out_features, translated.out_features)
        self.assertIsNotNone(translated.bias)
        self.assertEqual(str(target.weight_quant), str(translated.weight_quant))
        self.assertEqual(str(target.bias_quant), str(translated.bias_quant))

    def test_translate_linear_layer_ternary_weight_bias_quant(self):
        layer = self.create_qtorch_linear_layer(quantizer=Ternarize())
        target = self.create_brevitas_linear_layer(
            weight_quant=bquant.TernaryWeights, bias_quant=bquant.TernaryBias
        )
        translated = translate_linear_layer(layer)

        self.assertIsInstance(translated, bnn.QuantLinear)
        self.assertEqual(target.in_features, translated.in_features)
        self.assertEqual(target.out_features, translated.out_features)
        self.assertIsNotNone(translated.bias)
        self.assertEqual(str(target.weight_quant), str(translated.weight_quant))
        self.assertEqual(str(target.bias_quant), str(translated.bias_quant))

    def test_translate_linear_layer_on_random_input(self):
        layer = self.create_qtorch_linear_layer(quantizer=Ternarize())

        qtorch_model = nn.Sequential(layer)

        translated_linear = translate_linear_layer(layer)
        translated_linear.weight = nn.Parameter(layer.weight)
        translated_linear.bias = nn.Parameter(layer.bias)

        translated_model = nn.Sequential(translated_linear)

        random_input = torch.randn((1, 5))
        qtorch_output = qtorch_model(random_input)
        translated_output = translated_model(random_input)

        self.assertListEqual(qtorch_output.tolist(), translated_output.tolist())


if __name__ == "__main__":
    unittest.main()



import unittest

import torch
from torch import nn
import random
import numpy as np

from brevitas import quant
import brevitas.nn as bnn

from elasticai.creator.brevitas.brevitas_model_comparison import (
    BrevitasModelComparisonTestCase,
)
import elasticai.creator.brevitas.brevitas_quantizers as bquant


class TestModelComparison(BrevitasModelComparisonTestCase):
    """
    Test the brevitas model comparison class
    """

    def setUp(self) -> None:
        self.ensure_reproducibility()

    @staticmethod
    def ensure_reproducibility():
        torch.manual_seed(0)
        random.seed(0)
        np.random.seed(0)

    @classmethod
    def modelA(cls):
        l1 = bnn.QuantLinear(in_features=5, out_features=2, bias=True)
        # we think brevitas is manipulating some seed therefore we need to reset them again after calling quantconv1d
        cls.ensure_reproducibility()
        return nn.Sequential(l1)

    @staticmethod
    def modelB():
        l1 = bnn.QuantConv1d(in_channels=5, out_channels=2, kernel_size=3)
        return nn.Sequential(l1)

    @staticmethod
    def modelC():
        l1 = bnn.QuantLinear(in_features=5, out_features=2, bias=True)
        l2 = bnn.QuantConv1d(in_channels=1, out_channels=1, kernel_size=2)
        return nn.Sequential(l1, l2)

    @classmethod
    def big_model(cls):
        l1 = bnn.QuantConv1d(
            in_channels=3,
            out_channels=3,
            kernel_size=2,
            weight_quant=bquant.BinaryWeights,
            bias_quant=bquant.BinaryBias,
        )
        l2 = nn.BatchNorm1d(num_features=3)
        l3 = bnn.QuantIdentity(act_quant=bquant.BinaryActivation)
        l4 = nn.MaxPool1d(
            kernel_size=3,
            stride=2,
        )
        l5 = nn.Flatten()
        l6 = bnn.QuantLinear(
            in_features=21, out_features=1, bias=True, weight_quant=bquant.BinaryWeights
        )
        l7 = nn.Sigmoid()

        # we think brevitas is manipulating some seed therefore we need to reset them again after calling quantconv1d
        cls.ensure_reproducibility()

        return nn.Sequential(l1, l2, l3, l4, l5, l6, l7)

    def test_same_easy_model(self):
        self.assertModelEqual(self.modelA(), self.modelA())

    def test_same_big_model(self):
        self.assertModelEqual(self.big_model(), self.big_model())

    def test_different_model_with_same_length(self):
        with self.assertRaises(AssertionError):
            self.assertModelEqual(self.modelA(), self.modelB())

    def test_different_model_with_different_length(self):
        with self.assertRaises(AssertionError):
            self.assertModelEqual(self.modelA(), self.modelC())

    def test_model_with_different_quantizers(self):
        model1 = nn.Sequential(
            bnn.QuantLinear(
                in_features=5,
                out_features=2,
                bias=True,
            )
        )

        model2 = nn.Sequential(
            bnn.QuantLinear(
                in_features=5,
                out_features=2,
                bias=True,
                weight_quant=quant.SignedBinaryWeightPerTensorConst,
            )
        )

        with self.assertRaises(AssertionError):
            self.assertModelEqual(model1, model2)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['elasticai',
 'elasticai.creator',
 'elasticai.creator.brevitas',
 'elasticai.creator.brevitas.translation_functions',
 'elasticai.creator.examples',
 'elasticai.creator.integrationTests',
 'elasticai.creator.systemTests',
 'elasticai.creator.tests',
 'elasticai.creator.tests.vhdl',
 'elasticai.creator.vhdl',
 'elasticai.creator.vhdl.generator',
 'elasticai.creator.vhdl.generator.functions']

package_data = \
{'': ['*'],
 'elasticai.creator.tests.vhdl': ['vhdFiles/*'],
 'elasticai.creator.vhdl': ['source/*', 'testbench/*']}

install_requires = \
['MonkeyType>=21.5.0,<22.0.0', 'torch>=1.10.0,<2.0.0']

extras_require = \
{'brevitas': ['brevitas>=0.7.0,<0.8.0',
              'onnx>=1.10.2,<2.0.0',
              'onnxoptimizer>=0.2.6,<0.3.0'],
 'examples': ['torchvision>=0.11.2,<0.12.0'],
 'systemtests': ['torchvision>=0.11.2,<0.12.0']}

setup_kwargs = {
    'name': 'elasticai.creator',
    'version': '0.3.3a2',
    'description': 'Design, train and compile neural networks optimized specifically for FPGAs.',
    'long_description': '# Elastic.ai Creator\n\nDesign, train and compile neural networks optimized specifically for FPGAs.\nObtaining a final model is typically a three stage process.\n* design and train it using the layers provided in the `elasticai.creator` package.\n* translate the model to a target representation, e.g. VHDL\n* compile the intermediate representation with a third party tool, e.g. Xilinx Vivado (TM)\n\nThis version currently only supports [Brevitas](https://github.com/Xilinx/brevitas) as a target representation.\n\nThe project is part of the elastic ai ecosystem developed by the Embedded Systems Department of the University Duisburg-Essen. For more details checkout the slides at https://www.researchgate.net/publication/356372207_In-Situ_Artificial_Intelligence_for_Self-_Devices_The_Elastic_AI_Ecosystem_Tutorial.\n\n\n## Table of contents\n\n[[_TOC_]]\n\n## Requirements\n\nThe dependencies of the PyTorch implementation of MTLTranslator are listed below:\n- torch\n- torchvision\n- torchaudio\n- matplotlib\n- numpy\n- scikit-learn\n- brevitas\n- onnx\n- onnxoptimizer\n\n### Poetry\n\n#### Using ElasticAICreator as a dependency\nAdd the following to your `pyproject.toml`\n```toml\n[[tool.poetry.source]]\nname = "elasticai"\nurl = "https://git@git.uni-due.de/api/v4/projects/5522/packages/pypi/simple"\nsecondary = true\n```\nAfterwards you should be able to add the ElasticAICreator to your project like so\n```bash\npoetry add elasticaicreator\n```\n\n#### Developing the ElasticAICreator\nFor easy creating a virtual environment and having a guaranteed working environment you can use the\n[poetry](https://python-poetry.org/) module.\n\n(before installing poetry make sure you have python3.9 installed on your system)\n\npoetry can be installed in the following way:\n```bash\npip install poetry\n```\nif your default python version is not (3.9.*) you need to run the following command first:\n```bash\npoetry env use python3.9\n```\n\nAfter installing poetry you can create an environment and pull all necessary dependencies by just typing the following\ncommand in the project root folder where the ```poetry.lock``` and ```pyproject.toml``` is located:\n\n```bash\npoetry install\n```\n\nThe installed poetry environment can be activated typing the following command in the project root folder:\n```bash\npoetry shell\n```\nYou may consider installing poetry plugin in pycharm and adding the created environment\n\n## Installing\n\nIf you want to use the elastic.ai Creator Translator in your project you can install it via pip. \nTo install from git run the following command: \n```bash\npip install git+https://git.uni-due.de/embedded-systems/artificial-intelligence/toolchain/mtltranslator\n```\nTo build locally in the directory run:\n```bash\npip install  .\n```\n\n## Structure of the Project\n\nThe structure of the project is as follows.\nThe [qtorch](elasticai/creator/qtorch) folder includes our implementation of quantized PyTorch layer.\nThe folder [protocols](elasticai/creator/protocols) contains some general protocols for the models and layers which are also used by multiple translated languages.\nIn the [translator](elasticai/creator/translator) folder there are the modules which can be used for every translation from a pytorch model to a target language.\nThe subfolder [brevitas](elasticai/creator/brevitas) is for the translation to Brevitas.\nEach language we can translate to has folders for unit tests, integration tests and system test. \n\n## QTorch\n\nQTorch is a library of quantized PyTorch layers. \nThe [layers](elasticai/creator/layers.py) file includes all implemented quantized PyTorch layers.\nThese layers are the followings:\n- QConv1d for quantized Conv1d\n- QConv2d for quantized Conv2d\n- QLinear for quantized Linear\n- QLSTMCell for quantized LSTMCell\n- QLSTM for stacking QLSTMCell\n\nFor the quantization we implemented quantizers:\n- Binarize which converts the weights to be -1 or 1\n- Ternarize which converts the weights to be -1 or 0 or 1\n- QuantizeTwoBit use Residual Quantization to quantize to two bits\n- ResidualQuantization converts weights to a bit vector using the residual multilevel binarization method [RebNet, Ghasemzadeh et al. 2018](https://arxiv.org/pdf/1711.01243.pdf)\n\n\nWe wrote tests for the layers which can be found in the [layers_tests](elasticai/creator/qtorch/tests/test_layer.py).\nTo add constraints on the convolutional and linear layers you can use the [constraints](elasticai/creator/constraints.py) and can easily expand it with more constraints.\nWe also implemented blocks of convolution and linear layers consisting of the convolution or linear following a batch normalisation and some activation. \nAlso consider the [tests](elasticai/creator/qtorch/tests/test_block.py) for the blocks.\nFor getting more insight in the QTorch library consider the examples in the [examples](elasticai/creator/examples) folder.\n\n## Users guide \n\nAs a user one want to convert an existing pytorch model to one of our languages.\n1. Add our translator as a dependency in your project.\n2. Instantiate the model. \n3. Optionally you can train it or load some weights. \n4. Input the model in the translation function like shown in the following. \n\nPlease refer to the system test of each language as an example.\n\n## Translating to Brevitas\n\nHow to translate a given PyTorch model consisting of QTorch layers to Brevitas?\nThis is how to translate a given model to a Brevitas model:\n\n```python\nfrom elasticai.creator.brevitas.brevitas_representation import BrevitasRepresentation\n\nconverted_model = BrevitasRepresentation.from_pytorch(qtorch_model).translated_model\n```\nargs:\n- qtorch_model: a pytorch model (supports most of the [QTorch](https://git.uni-due.de/embedded-systems/artificial-intelligence/toolchain/qtorch) layers and some standard pytorch layers)\n\nreturns:\n- converted_model: a Brevitas model\n\nExample usages are shown here: [Brevitas system tests](elasticai/creator/systemTests).\nWe also support to translate a brevitas model to onnx which is shown in the system test.\n\n### Translations\n\nThe following QTorch or PyTorch layers are translated to the corresponding Brevitas layers:\n\n- QConv1d to QuantConv1d\n- QConv2d to QuantConv2d\n- QLinear to QuantLinear\n- Binarize to QuantIdentity\n- Ternarize to QuantIdentity\n- PyTorch MaxPool1d to PyTorch MaxPool1d\n- PyTorch BatchNorm1d to PyTorch BatchNorm1d \n- PyTorch Flatten to PyTorch Flatten\n- PyTorch Sigmoid to PyTorch Sigmoid\n\nYou can find the mappings in [translation_mapping](elasticai/creator/brevitas/translation_mapping.py) and can easily add more PyTorch layers.\n\n### Supported Layers for Brevitas Translation\n\n- QuantConv1d: quantized 1d convolution with weight- and bias-quantization\n- QuantConv2d: quantized 2d convolution with weight- and bias-quantization\n- QuantLinear: quantized linear layer with weight- and bias-quantization\n- QuantIdentity(act_quant=quantizers.BinaryActivation): binary activation layer\n- QuantIdentity(act_quant=quantizers.TernaryActivation): ternary activation layer\n\n### Limitations for Brevitas Translation\n\n- we do not support all QTorch layers in the QTorch repository. Not supported layers are:\n  - Ternarization with more complex thresholds e.g threshold of 0.1 \n  - ResidualQuantization \n  - QuantizeTwoBit \n  - QLSTM\n- we do not support all PyTorch layers, but you can easily add them in the [translation_mapping](elasticai/creator/brevitas/translation_mapping.py).\n\n## General Limitations\n\nBy now we only support Sequential models for our translations.\n\n## Developers introductory Guide and Glossary\n\nNew translation targets should be located in their own folder, e.g. Brevitas for translating from any language to Brevitas.\nWorkflow for adding a new translation:\n1. Obtain a structure, such as a list in a sequential case, which will describe the connection between every component.\n2. Identify and label relevant structures, in the base cases it can be simply separate layers.\n3. Map each structure to its function which will convert it, like for [example](elasticai/creator/brevitas/translation_mapping.py).\n4. Do such conversions.\n5. Recreate connections based on 1.\n\nEach sub-step should be separable and it helps for testing if common functions are wrapped around an adapter.\n\n## Model reporter\nAs part of this repository an utility called [model reporter](elasticai/creator/model_reporter.py) exists which is used to produce a file with the individual identifier from a given dataset. \nIt is used for hardware comparisons, where a small subset of the data is used for comparison. \nExample: \n```\necg-path ground-truth model-prediction\nsinus/08219_107.ecg 0 0 \natrial_fibrillation/07879_392.ecg 1 1\nsinus/08219_573.ecg 0 0\nsinus/04126_477.ecg 0 0\natrial_fibrillation/07162_287.ecg 1 1\natrial_fibrillation/06426_789.ecg 1 1\natrial_fibrillation/05121_664.ecg 1 1\n```\nThis is expected to use a path to help a validator to automatically extract the data from the said path and add the further predictions.\nExample usage:\n```python\nreport = ModelReport(model=loaded_model, data=[labels,inputs,data[1]], is_binary=True, threshold=0.5)\nreport.write_to_csv(path="results.csv")\n```\n\n## Tests\n\nOur implementation is fully tested with unit, integration and system tests.\nPlease refer to the system tests as examples of how to use the Elastic Ai Creator Translator.\nYou can run one explicit test with the following statement: \n\n```python -m unittest discover -p "test_*.py" elasticai/creator/translator/path/to/test.py```\n\nIf you want to run all unit tests for example, give the path to the unit tests:\n\n```python -m unittest discover -p "test_*.py" elasticai/creator/translator/path/to/language/unitTests/```\n\nYou can also run all tests together:\n\n```python -m unittest discover -p "test_*.py" elasticai/creator/translator/path/to/language/```\n\nIf you want to add more tests please refer to the [Test Guidelines](test_guidelines.md).\n\n### Brevitas System Tests\n\nThe [Brevitas system tests](elasticai/creator/systemTests) can be used as example use cases of our implementation.\nWe created tests which check the conversion of a model like we would expect our models will look like.\nIn addition, we also created tests for validating the conversion for trained models or unusual models. \nNote that you have to use your own data set and therefore maybe do some small adaptions by using the training.\n\n### Brevitas Integration Tests\n\nOur  [Brevitas integration tests](elasticai/creator/integrationTests) are focused on testing the conversion of one specific layer. \nWe created for all our supported layers a minimal model with this layer included and test its functionality. \n\n### Brevitas Unit Tests\n\nIn addition to system and integration tests we implemented unit tests. \nThe unit tests of each module is named like the model but starting with "test_" and can be found in the unitTest folder.\nThe Brevitas unit tests can be found [here](elasticai/creator/brevitas/unitTests).\n\n## VHDL Generation\n\nWe follow the VHDL code guidelines of this [book](https://github.com/fabriziotappero/Free-Range-VHDL-book).\n\n## Developers Guide\n* We use [black](https://black.readthedocs.io/en/stable/index.html) for code formatting. For instruction on setup with your IDE please refer to https://black.readthedocs.io/en/stable/integrations/editors.html#editor-integration.\n* importing `*` should be avoided in favor of explicit imports\n\n### Test style Guidelines\n\n#### File IO\nIn general try to avoid interaction with the filesystem. In most cases instead of writing to or reading from a file you can use a StringIO object or a StringReader.\nIf you absolutely have to create files, be sure to use pythons [tempfile](https://docs.python.org/3.9/library/tempfile.html) module and cleanup after the tests.\n\n\n#### Diectory structure and file names\nFiles containing tests for a python module should be located in a test directory for the sake of separation of concerns. \nEach file in the test directory should contain tests for one and only one class/function defined in the module. \nFiles containing tests should be named according to the rubric\n`test_ClassName.py`.\nNext, if needed for more specific tests define a class which is a subclass of unittest.TestCase like [test_brevitas_model_comparison](elasticai/creator/integrationTests/test_brevitas_model_comparison.py) in the integration tests folder. \nThen subclass it, in this class define a setUp method (and possibly tearDown) to create the global environment. \nIt avoids introducing the category of bugs associated with copying and pasting code for reuse. \nThis class should be named similarly to the file name.\nThere\'s a category of bugs that appear if  the initialization parameters defined at the top of the test file are directly used: some tests require the initialization parameters to be changed slightly. \nIts possible to define a parameter and have it change in memory as a result of a test. \nSubsequent tests will therefore throw errors.\nEach class contains methods that implement a test. \nThese methods are named according to the rubric\n`test_name_condition`\n\n#### Unit tests\nIn those tests each functionality of each function in the module is tested, it is the entry point  when adding new functions. \nIt assures that the function behaves correctly independently of others. \nEach test has to be fast, so use of heavier libraries is discouraged.\nThe input used is the minimal one needed to obtain a reproducible output. \nDependencies should be replaced with mocks as needed. \n\n#### Integration Tests\nHere the functions\' behaviour with other modules is tested. \nIn this repository each integration function is in the correspondent folder.\nThen the integration with a single class of the target, or the minimum amount of classes for a functionality, is tested in each separated file.\n\n#### System tests\nThose tests will use every component of the system, comprising multiple classes.\nThose tests include expected use cases and unexpected or stress tests.\n\n#### Adding new functionalities and tests required\nWhen adding new functions to an existing module, add unit tests in the correspondent file in the same order of the module, if a new module is created a new file should be created.\nWhen a bug is solved created the respective regression test to ensure that it will not return.\nProceed similarly with integration tests. \nCreating a new file if a functionality completely different from the others is created e.g. support for a new layer.\nSystem tests are added if support for a new library is added.\n\n#### Updating tests\nIf new functionalities are changed or removed the tests are expected to reflect that, generally the ordering is unit tests -> integration tests-> system tests.\nAlso, unit tests that change the dependencies should be checked, since this system is fairly small the internal dependencies are not always mocked.\n\nreferences: https://jrsmith3.github.io/python-testing-style-guidelines.html',
    'author': 'Department Embedded Systems - University Duisburg Essen',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/es-ude/elastic-ai.creator',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<3.10',
}


setup(**setup_kwargs)



import unittest
from elasticai.creator.vhdl.generator.testbench_strings import (
    get_type_definitions_string,
    get_clock_process_string,
    get_uut_string,
    get_test_process_header_string,
    get_test_process_end_string,
)


class TestbenchStringsTest(unittest.TestCase):
    def test_generate_type_definition(self) -> None:
        expected_inputs_lines = [
            "    ------------------------------------------------------------",
            "    -- Testbench Data Type",
            "    ------------------------------------------------------------",
            "    type RAM_ARRAY is array (0 to 9 ) of signed(DATA_WIDTH-1 downto 0);",
        ]
        type_definition_string = get_type_definitions_string(
            type_dict={
                "RAM_ARRAY": "array (0 to 9 ) of signed(DATA_WIDTH-1 downto 0)",
            }
        )
        for i in range(len(expected_inputs_lines)):
            self.assertEqual(
                expected_inputs_lines[i], type_definition_string.splitlines()[i]
            )

    def test_generate_clock_process(self) -> None:
        expected_clock_lines = [
            "    clock_process : process",
            "    begin",
            "        clk <= '0';",
            "        wait for clk_period/2;",
            "        clk <= '1';",
            "        wait for clk_period/2;",
            "    end process; -- clock_process",
        ]
        clock_process_string = get_clock_process_string()
        for i in range(len(expected_clock_lines)):
            self.assertEqual(
                expected_clock_lines[i], clock_process_string.splitlines()[i]
            )

    def test_generate_uut(self) -> None:
        expected_uut_lines = [
            "    uut: sigmoid",
            "    port map (",
            "        x => test_input,",
            "        y => test_output",
            "    );",
        ]
        utt_string = get_uut_string(
            component_name="sigmoid",
            mapping_dict={"x": "test_input", "y": "test_output"},
        )
        for i in range(len(expected_uut_lines)):
            self.assertEqual(expected_uut_lines[i], utt_string.splitlines()[i])

    def test_generate_bigger_uut(self) -> None:
        expected_uut_lines = [
            "    uut: lstm_common_gate",
            "    port map (",
            "        reset => reset,",
            "        clk => clock,",
            "        x => x,",
            "        w => w,",
            "        b => b,",
            "        vector_len => vector_len,",
            "        idx => idx,",
            "        ready => ready,",
            "        y => y",
            "    );",
        ]
        utt_string = get_uut_string(
            component_name="lstm_common_gate",
            mapping_dict={
                "reset": "reset",
                "clk": "clock",
                "x": "x",
                "w": "w",
                "b": "b",
                "vector_len": "vector_len",
                "idx": "idx",
                "ready": "ready",
                "y": "y",
            },
        )
        for i in range(len(expected_uut_lines)):
            self.assertEqual(expected_uut_lines[i], utt_string.splitlines()[i])

    def test_generate_test_process_header(self) -> None:
        expected_test_process_header_lines = [
            "    test_process: process is",
            "    begin",
            '        Report "======Simulation start======" severity Note;',
        ]
        test_process_header_string = get_test_process_header_string()
        for i in range(len(expected_test_process_header_lines)):
            self.assertEqual(
                expected_test_process_header_lines[i],
                test_process_header_string.splitlines()[i],
            )

    def test_generate_test_process_end(self) -> None:
        expected_test_process_end_lines = [
            "",
            "        -- if there is no error message, that means all test case are passed.",
            '        report "======Simulation Success======" severity Note;',
            '        report "Please check the output message." severity Note;',
            "",
            "        -- wait forever",
            "        wait;",
            "",
            "    end process; -- test_process",
        ]
        test_process_end_string = get_test_process_end_string()
        for i in range(len(expected_test_process_end_lines)):
            self.assertEqual(
                expected_test_process_end_lines[i],
                test_process_end_string.splitlines()[i],
            )



import brevitas.nn as bnn

from torch.nn import Module

from elasticai.creator.layers import QLinear
from elasticai.creator.brevitas.translation_functions.translation_function_tools import (
    set_quantizers,
)


def translate_linear_layer(layer: QLinear) -> bnn.QuantLinear:
    """
    translates a qtorch linear layer into a brevitas linear layer
    the arguments convertion of the quantizers is done in an extra file because it is also used by the convolutional layers
    Args:
        layer (Linear): qtorch layer
    Returns:
        brevitas layer
    """

    args = {
        "in_features": layer.in_features,
        "out_features": layer.out_features,
        "bias": layer.bias is not None,
    }

    set_quantizers(layer, args)

    return bnn.QuantLinear(**args)



from typing import Callable

from elasticai.creator.brevitas.translation_functions import (
    translate_conv1d,
    translate_conv2d,
    translate_linear_layer,
    translate_binarize_layer,
    translate_ternarize_layer,
    translate_layer,
)
from torch.nn import Module


class ConversionMapping:
    """
    Class to store all currently possible qtorch->brevitas mappings.
    """

    def __init__(self):
        self._mapping = {
            "Binarize": translate_binarize_layer,
            "Ternarize": translate_ternarize_layer,
            "ParametrizedQLinear": translate_linear_layer,
            "ParametrizedQConv1d": translate_conv1d,
            "ParametrizedQConv2d": translate_conv2d,
            "MaxPool1d": translate_layer,
            "BatchNorm1d": translate_layer,
            "Flatten": translate_layer,
            "Sigmoid": translate_layer,
        }

    def get_conversion_function(self, layer: Module) -> Callable[[Module], Module]:
        """
        return conversion function for passed layer
        Args:
            layer (Layer): layer which should be converted
        Returns:
            conversion function for layer
        """
        layer_name = layer.__class__.__name__
        try:
            return self._mapping[layer_name]
        except KeyError:
            raise NotImplementedError(
                f"For the layer {layer_name} no mapping to a conversion function is implemented."
            )



from typing import Any, TypeVar, Generic

T = TypeVar("T")


class HasTag(Generic[T]):
    def elasticai_tags(self) -> dict:
        pass


def tag(module: T, **new_tags: Any) -> HasTag:
    """Add tags to any object wrapping it in a TagWrapper if necessary

    new_tags will override possibly existing tags
    """
    old_tags = get_tags(module)
    tags = old_tags | new_tags
    module._elasticai_tags = tags

    def get_tags_method(self=module) -> dict:
        return self._elasticai_tags

    module.elasticai_tags = get_tags_method
    return module


def get_tags(module: Any) -> dict:
    if hasattr(module, "elasticai_tags"):
        return module.elasticai_tags()
    return {}


def has_tag(module: Any, tag_name) -> bool:
    return tag_name in get_tags(module)



from elasticai.creator.vhdl.generator.generator_functions import (
    precomputed_scalar_function_process,
)
from elasticai.creator.tests.vhdl.vhdl_file_testcase import GeneratedVHDLCodeTest


class PrecomputedScalarFunctionProcessTest(GeneratedVHDLCodeTest):
    def test_empty_x_list_y_list_one_element(self) -> None:
        code = precomputed_scalar_function_process(x_list=[], y_list=[1])
        self.assertEqual(code, 'y <= "0000000100000000";' + "\n" + "\t" + "\t")

    def test_empty_x_list_y_list_too_many_elements(self) -> None:
        self.assertRaises(
            ValueError, precomputed_scalar_function_process, x_list=[], y_list=[1, 2]
        )

    def test_empty_y_list(self) -> None:
        self.assertRaises(
            ValueError, precomputed_scalar_function_process, x_list=[], y_list=[]
        )

    def test_x_list_lengths_not_suitable_for_y_list_lengths(self) -> None:
        self.assertRaises(
            ValueError, precomputed_scalar_function_process, x_list=[1, 2], y_list=[1]
        )

    def test_x_list_with_only_one_element(self) -> None:
        expected_code = """        if int_x<256 then
                y <= "0000000100000000"; -- 256
            else
                y <= "0000001000000000"; -- 512
            end if;"""
        code = precomputed_scalar_function_process(x_list=[1], y_list=[1, 2])
        self.check_generated_code(expected_code, code)

    def test_unsorted_x_list(self) -> None:
        expected_code = """        if int_x<256 then
                    y <= "0000000100000000"; -- 256
                elsif int_x<512 then
                    y <= "0000001000000000"; -- 512
                elsif int_x<768 then
                    y <= "0000001100000000"; -- 768
                else
                    y <= "0000010000000000"; -- 1024
                end if;"""
        code = precomputed_scalar_function_process(
            x_list=[3, 1, 2], y_list=[1, 2, 3, 4]
        )
        self.check_generated_code(expected_code, code)



import csv
from typing import Sequence, Iterable, Any
import torch

from elasticai.creator.protocols import Model


class ModelReport:
    """
    The `ModelReport` aims to ease the verification of neural networks
    implemented in hardware. Given a model and a corresponding data
    provider it will produce a csv file containing the model's inference
    results on said data. Additionally it will store the paths to the
    inferred data samples so that another tool can use the data samples
    and the produced csv file to check that the hw model produces the
    same output as the software model for every data sample.
    """

    def __init__(
        self,
        model: Model,
        data: Iterable[Sequence[str, Any, Any]],
        is_binary: bool,
        threshold: float = 0.5,
    ) -> None:
        model.eval()
        self.model = model
        self.data: Iterable[Sequence[str, Any, Any]] = data
        self.evaluate(is_binary, threshold)

    def evaluate(self, is_binary: bool, threshold: float = 0.5) -> None:
        # TODO: add support for more labels
        """
        Create a results tuple from the input model and data, it is designed to be used in classification
        Args:
            is_binary (bool): Binary labels are handled slightly differently so it needs a bool value, currently only supported if True
            threshold (float): If binary the threshold to compare the model to
        """
        if is_binary:
            self._results = [
                (
                    example_id,
                    int(expected_value),
                    int(self.model(torch.unsqueeze(input_value, 1)) > threshold),
                )
                for example_id, input_value, expected_value in zip(
                    self.data[0], self.data[1], self.data[2]
                )
            ]

    def write_to_csv(self, path: str) -> None:
        """
        Writes the results to the path in a csv format
        Args:
            path(str): the results path eg "results.csv"
        """
        with open(path, "wt") as results_file:
            wtr = csv.writer(results_file, delimiter=" ", lineterminator="\n")
            wtr.writerow(["path", "ground-truth", "model-prediction"])
            wtr.writerows(self._results)

    def write_to_txt(self, path: str) -> None:
        """
        Writes the results to the path in a txt format
        Args:
            path(str): the results path eg "results.txt"
        """
        with open(path, "wt") as results_file:
            results_file.write("path ground-truth model-prediction\n")
            for result in self._results:
                results_file.write("{} {} {}\n".format(result[0], result[1], result[2]))



from elasticai.creator.vhdl.generator.general_strings import (
    get_libraries_string,
    get_entity_or_component_string,
    get_architecture_header_string,
    get_signal_definitions_string,
    get_architecture_begin_string,
    get_variable_definitions_string,
    get_architecture_end_string,
)
from elasticai.creator.vhdl.generator.generator_functions import (
    generate_signal_definitions_for_lstm,
)
from elasticai.creator.vhdl.generator.vhd_strings import (
    get_file_path_string,
    get_gate_definition_string,
    get_port_map_string,
    get_define_process_string,
)

component_name = "lstm_cell"
file_name = component_name + ".vhd"
architecture_name = "lstm_cell_rtl"

DATA_WIDTH = 16
FRAC_WIDTH = 8


def main():
    file_path = get_file_path_string(folder_names=["..", "source"], file_name=file_name)

    with open(file_path, "w") as writer:
        writer.write(get_libraries_string(work_lib=True))
        writer.write(
            get_entity_or_component_string(
                entity_or_component="entity",
                entity_or_component_name=component_name,
                data_width=DATA_WIDTH,
                frac_width=FRAC_WIDTH,
                variables_dict={
                    "x": "in signed(DATA_WIDTH-1 downto 0)",
                    "c_in": "in signed(DATA_WIDTH-1 downto 0)",
                    "h_in": "in signed(DATA_WIDTH-1 downto 0)",
                    "c_out": "out signed(DATA_WIDTH-1 downto 0)",
                    "h_out": "out signed(DATA_WIDTH-1 downto 0)",
                },
            )
        )

        # architecture structure
        writer.write(
            get_architecture_header_string(
                architecture_name=architecture_name, component_name=component_name
            )
        )

        # generate lstm signal definitions
        lstm_signal_definitions = generate_signal_definitions_for_lstm(
            data_width=DATA_WIDTH, frac_width=FRAC_WIDTH
        )
        writer.write(get_signal_definitions_string(lstm_signal_definitions))

        # string of input/forget/cell/output/ gate definition or new cell state definition
        writer.write(
            get_gate_definition_string(
                comment="-- Intermediate results\n"
                "-- Input gate without/with activation\n"
                "-- i = \sigma(W_{ii} x + b_{ii} + W_{hi} h + b_{hi})",
                signal_names=["i_wo_activation", "i"],
            )
        )
        writer.write(
            get_gate_definition_string(
                comment="-- Forget gate without/with activation\n"
                "-- f = \sigma(W_{if} x + b_{if} + W_{hf} h + b_{hf})",
                signal_names=["f_wo_activation", "f"],
            )
        )
        writer.write(
            get_gate_definition_string(
                comment="-- Cell gate without/with activation\n"
                "-- g = \\tanh(W_{ig} x + b_{ig} + W_{hg} h + b_{hg})",
                signal_names=["g_wo_activation", "g"],
            )
        )
        writer.write(
            get_gate_definition_string(
                comment="-- Output gate without/with activation\n"
                "-- o = \sigma(W_{io} x + b_{io} + W_{ho} h + b_{ho})",
                signal_names=["o_wo_activation", "o"],
            )
        )
        writer.write(
            get_gate_definition_string(
                comment="-- new_cell_state without/with activation\n"
                "-- c' = f * c + i * g",
                signal_names=["c_new", "c_new_wo_activation"],
            )
        )
        writer.write(
            get_gate_definition_string(
                comment="""-- h' = o * \\tanh(c')""", signal_names=["h_new"]
            )
        )
        # components
        writer.write(
            get_entity_or_component_string(
                entity_or_component="component",
                entity_or_component_name="mac_async",
                data_width="DATA_WIDTH",
                frac_width="FRAC_WIDTH",
                variables_dict={
                    "x1": "in signed(DATA_WIDTH-1 downto 0)",
                    "x2": "in signed(DATA_WIDTH-1 downto 0)",
                    "w1": "in signed(DATA_WIDTH-1 downto 0)",
                    "w2": "in signed(DATA_WIDTH-1 downto 0)",
                    "b": "in signed(DATA_WIDTH-1 downto 0)",
                    "y": "out signed(DATA_WIDTH-1 downto 0)",
                },
            )
        )
        writer.write(
            get_entity_or_component_string(
                entity_or_component="component",
                entity_or_component_name="sigmoid",
                data_width="DATA_WIDTH",
                frac_width="FRAC_WIDTH",
                variables_dict={
                    "x": "in signed(DATA_WIDTH-1 downto 0)",
                    "y": "out signed(DATA_WIDTH-1 downto 0)",
                },
            )
        )
        writer.write(
            get_entity_or_component_string(
                entity_or_component="component",
                entity_or_component_name="tanh",
                data_width="DATA_WIDTH",
                frac_width="FRAC_WIDTH",
                variables_dict={
                    "x": "in signed(DATA_WIDTH-1 downto 0)",
                    "y": "out signed(DATA_WIDTH-1 downto 0)",
                },
            )
        )
        # architecture behavior
        writer.write(get_architecture_begin_string())
        writer.write(
            get_variable_definitions_string(
                {"c_out": "c_new_wo_activation", "h_out": "h_new"}
            )
        )
        # port mapping
        writer.write(
            get_port_map_string(
                map_name="FORGET_GATE_MAC",
                component_name="mac_async",
                signals={
                    "x1": "x",
                    "x2": "h_in",
                    "w1": "wif",
                    "w2": "whf",
                    "b": "bf",
                    "y": "f_wo_activation",
                },
            )
        )
        writer.write(
            get_port_map_string(
                map_name="FORGET_GATE_SIGMOID",
                component_name="sigmoid",
                signals=["f_wo_activation", "f"],
            )
        )
        writer.write(
            get_port_map_string(
                map_name="INPUT_GATE_MAC",
                component_name="mac_async",
                signals={
                    "x1": "x",
                    "x2": "h_in",
                    "w1": "wii",
                    "w2": "whi",
                    "b": "bi",
                    "y": "i_wo_activation",
                },
            )
        )
        writer.write(
            get_port_map_string(
                map_name="INPUT_GATE_SIGMOID",
                component_name="sigmoid",
                signals=["i_wo_activation", "i"],
            )
        )
        writer.write(
            get_port_map_string(
                map_name="CELL_GATE_MAC",
                component_name="mac_async",
                signals={
                    "x1": "x",
                    "x2": "h_in",
                    "w1": "wig",
                    "w2": "whg",
                    "b": "bg",
                    "y": "g_wo_activation",
                },
            )
        )
        writer.write(
            get_port_map_string(
                map_name="CELL_GATE_TANH",
                component_name="tanh",
                signals=["g_wo_activation", "g"],
            )
        )
        writer.write(
            get_port_map_string(
                map_name="NEW_CELL_STATE_MAC",
                component_name="mac_async",
                signals={
                    "x1": "f",
                    "x2": "i",
                    "w1": "c_in",
                    "w2": "g",
                    "b": "(others=>'0')",
                    "y": "c_new_wo_activation",
                },
            )
        )
        writer.write(
            get_port_map_string(
                map_name="NEW_CELL_STATE_TANH",
                component_name="tanh",
                signals=["c_new_wo_activation", "c_new"],
            )
        )
        writer.write(
            get_port_map_string(
                map_name="MAC_ASYNC_4",
                component_name="mac_async",
                signals={
                    "x1": "x",
                    "x2": "h_in",
                    "w1": "wio",
                    "w2": "who",
                    "b": "bo",
                    "y": "o_wo_activation",
                },
            )
        )
        writer.write(
            get_port_map_string(
                map_name="SIGMOID_1",
                component_name="sigmoid",
                signals={"x": "o_wo_activation", "y": "o"},
            )
        )
        # write process
        writer.write(
            get_define_process_string(
                process_name="H_OUT_PROCESS",
                sensitive_signals_list=["o", "c_new"],
                behavior="h_new <= shift_right((o*c_new), FRAC_WIDTH)(DATA_WIDTH-1 downto 0);",
            )
        )

        writer.write(get_architecture_end_string(architecture_name=architecture_name))


if __name__ == "__main__":
    main()



import unittest
from torch import nn


class BrevitasModelEqualsTest(unittest.TestCase):
    def test_two_empty_models_are_unequal(self):
        self.assertNotEqual(nn.Sequential(), nn.Sequential())


if __name__ == "__main__":
    unittest.main()
